# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import time
from collections import defaultdict
import os
from itertools import dropwhile, chain

from msprobe.core.common.file_utils import check_file_or_directory_path, save_json, make_dir, load_json
from msprobe.core.common.log import logger
from msprobe.core.common.const import Const
from msprobe.overflow_check.utils import (RankPath, FileCache, is_communication_op, is_ignore_op, OverFlowCheckConst,
                                          analyze_anomaly_in_group)
from msprobe.overflow_check.graph import DataNode, CommunicationNode


class OverFlowCheck:
    def __init__(self, input_path, output_path):
        self._input_path = input_path
        self._output_path = output_path
        self._paths = {}
        self._resolve_input_path()
        self._check_framework()
        self._anomaly_nodes = []  # 记录所有异常节点
        self._cache = FileCache()
        self._first_comm_nodes = {}  # 记录各rank下首个通信节点的node_id
        self._after_comm_anomalies = {}  # 记录各rank下发生在通信节点之后的异常计算节点
        self._rank_comm_nodes_dict = {}  # 记录各rank的通信节点

    def analyze(self):
        for analyze_func in [self._pre_analyze, self._analyze, self._post_analyze]:
            analyze_func()
            if self._anomaly_nodes:
                self._gen_analyze_info()
                return
        logger.info('Cannot find any anomaly node, no need to generate analyze file.')

    def _resolve_input_path(self):
        contents = os.listdir(self._input_path)
        for path in contents:
            if not path.startswith('rank'):
                continue
            rank_str = path[len('rank'):]
            if not rank_str:
                rank = 0
            elif not rank_str.isdigit():
                continue
            else:
                rank = int(rank_str)
            dump_path = os.path.join(self._input_path, path, OverFlowCheckConst.DUMP_FILE)
            construct_path = os.path.join(self._input_path, path, OverFlowCheckConst.CONSTRUCT_FILE)
            stack_path = os.path.join(self._input_path, path, OverFlowCheckConst.STACK_FILE)
            self._paths[rank] = RankPath(rank, dump_path, construct_path, stack_path)

    def _check_framework(self):
        for _, rankpath in self._paths.items():
            data = load_json(rankpath.dump_path)
            if data.get('framework', '') != "pytorch":
                logger.error('overflow check is only available for PyTorch, though the data is stored in '
                             'a different framework.')
                raise RuntimeError(f'invalid dump_path {rankpath.dump_path}')
            break

    def _pre_analyze(self):
        logger.info('Start searching anomaly node before communication.')
        for path in self._paths.values():
            dump_data = self._cache.load_json(path.dump_path).get('data')
            if not dump_data:
                logger.warning(f'Rank {path.rank} has no dump data!')
                continue
            for op_name, op_data in dump_data.items():
                if is_communication_op(op_name):
                    self._first_comm_nodes[path.rank] = op_name
                    break
                data_node = DataNode(op_name, path.rank, op_data)
                if data_node.is_anomaly():
                    self._anomaly_nodes.append(data_node)
                    break

    def _analyze(self):
        logger.info('Start searching anomaly node during communication.')
        self._rank_comm_nodes_dict = {rank: self._analyze_comm_nodes(rank) for rank in self._paths}
        self._connect_comm_nodes()
        self._pruning()
        self._search_first_anomaly()

    def _post_analyze(self):
        logger.info('Start searching anomaly node after communication.')
        for nodes in self._after_comm_anomalies.values():
            if nodes:
                self._anomaly_nodes.append(nodes[0])

    def _gen_analyze_info(self):
        if not os.path.exists(self._output_path):
            make_dir(self._output_path)
        file_name = f'anomaly_analyze_{time.time_ns()}.json'
        result_file = os.path.join(self._output_path, file_name)
        result_content = defaultdict(list)
        for node in self._anomaly_nodes:
            result_content[f'rank_{node.rank}'].append(node.gen_node_info(self._paths[node.rank]))
        save_json(result_file, result_content, 2)
        logger.info(f"The analyze result is saved in: {result_file}")

    def _analyze_comm_nodes(self, rank):
        path = self._paths[rank]
        data = self._cache.load_json(path.dump_path).get('data')
        communication_nodes = {}
        if rank not in self._first_comm_nodes:  # 此rank没有通信节点
            return communication_nodes
        last_node_id = None  # 记录上一个通信节点的node_id
        compute_ops = []  # 记录两个通信节点之间的计算节点
        sub_layer = 0  # 记录两个通信算子之间异常计算节点的调用序数
        for op_name in dropwhile(lambda k: k != self._first_comm_nodes[rank], data):
            node_id = f'{rank}.{op_name}'
            op_data = data[op_name]
            if is_communication_op(op_name):
                comm_node = CommunicationNode(node_id, rank, DataNode(op_name, rank, op_data, sub_layer=sub_layer),
                                              compute_ops=compute_ops)
                if last_node_id:
                    communication_nodes.get(last_node_id).add_next(comm_node)
                communication_nodes[node_id] = comm_node
                last_node_id = node_id
                compute_ops = []
                sub_layer = 0
            elif not is_ignore_op(op_name):
                data_node = DataNode(op_name, rank, op_data, sub_layer=sub_layer)
                if data_node.is_anomaly():
                    compute_ops.append(data_node)
                sub_layer += 1
        if compute_ops:
            self._after_comm_anomalies[rank] = compute_ops
        return communication_nodes

    def _connect_comm_nodes(self):
        searched_ranks = set()
        for rank, nodes in list(self._rank_comm_nodes_dict.items())[:-1]:
            searched_ranks.add(rank)
            seen_nodes = set()
            for cur_node in nodes.values():
                conn_info = cur_node.find_connected_nodes()
                if not conn_info.get('ranks'):
                    conn_info['ranks'] = self._rank_comm_nodes_dict.keys()
                if not self._find_connection(conn_info, cur_node, searched_ranks, seen_nodes):
                    logger.info(f'Cannot find connected communication node for "{cur_node.node_id}".')

    def _find_connection(self, conn_info, cur_node, searched_ranks, seen_nodes):
        def connect():
            seen_nodes.add(search_node.node_id)
            if search_node.type == OverFlowCheckConst.DST:
                cur_node.add_dst(search_node)
            elif search_node.type == OverFlowCheckConst.SRC:
                search_node.layer = cur_node.layer
                search_node.add_dst(cur_node)
            else:
                cur_node.add_link(search_node)

        found = cur_node.connected
        for connected_rank in conn_info['ranks']:
            if connected_rank in searched_ranks:
                continue
            tar_id_prefix = f'{connected_rank}.{conn_info["api"]}'
            for search_id, search_node in self._rank_comm_nodes_dict[connected_rank].items():
                if search_id in seen_nodes:
                    continue
                if not (search_id.startswith(tar_id_prefix) and search_node.type == conn_info.get('type')):
                    continue
                search_conn_ranks = search_node.find_connected_nodes().get('ranks')
                if ((not search_conn_ranks and search_node.api not in OverFlowCheckConst.DIRECTED_API) or
                    cur_node.rank in search_conn_ranks):  # 有些无向通信算子没有填ProcessGroup，默认连接所有rank
                    connect()
                    found = True
                    break
        return found

    def _pruning(self):
        deleted_node_id = []
        for nodes in self._rank_comm_nodes_dict.values():
            for node_id in list(nodes.keys()):
                node = nodes[node_id]
                if node.has_nan_inf() or node.compute_ops:
                    continue
                deleted_node_id.append(node_id)
                node.delete()
                del nodes[node_id]
        logger.debug(f'After pruning, following nodes are removed: [{", ".join(deleted_node_id)}]')

    def _search_first_anomaly(self):
        nodes_queues = []
        for comm_nodes in self._rank_comm_nodes_dict.values():
            nodes_queues.append(sorted(list(comm_nodes.values()), key=lambda x: x.layer))
        seen_nodes = set()

        def get_next_node(node_list):
            while node_list:
                next_node = node_list.pop(0)
                if next_node.node_id not in seen_nodes:
                    return next_node
            return None

        def find_all_members(ori_node):
            ids = get_relative_ids(ori_node)
            id_queue = list(chain(*[get_relative_ids(self._get_node_by_id(n_id)).difference(ids) for n_id in ids]))
            while id_queue:
                new_id = id_queue.pop(0)
                ids.add(new_id)
                id_queue.extend(get_relative_ids(self._get_node_by_id(new_id)).difference(ids))
            return ids

        def get_relative_ids(ori_node):
            if not ori_node:
                return set()
            return ({ori_node.node_id} | ori_node.link_nodes.keys() | ori_node.src_nodes.keys() |
                    ori_node.dst_nodes.keys())

        while any(nodes_queues):
            groups = []
            all_ids_in_groups = set()
            for nodes in nodes_queues:
                node = get_next_node(nodes)
                if not node:
                    continue
                if not groups or node.node_id in all_ids_in_groups:
                    new_group = find_all_members(node)
                    groups.append(new_group)
                    all_ids_in_groups.update(new_group)
            for group in groups:
                seen_nodes.update(group)
                self._anomaly_nodes.extend(analyze_anomaly_in_group([self._get_node_by_id(n_id) for n_id in group]))
            if self._anomaly_nodes:
                self._anomaly_nodes = [min(self._anomaly_nodes, key=lambda x: (x.layer, x.sub_layer))]
                return

    def _get_node_by_id(self, node_id):
        splits = node_id.split(Const.SEP, 1)
        if len(splits) < 2 or not splits[0].isdigit():
            logger.error(f'invalid node_id {node_id}')
            raise RuntimeError(f'invalid node_id {node_id}')
        rank = int(splits[0])
        return self._rank_comm_nodes_dict.get(rank, {}).get(node_id)


def _overflow_check_parser(parser):
    parser.add_argument("-i", "--input_path", dest="input_path", default="", type=str,
                        help="<Required> The dump file path, over step level. eg: \"xxx/step_0/\".",
                        required=True)
    parser.add_argument("-o", "--output_path", dest="output_path", default="./output", type=str,
                        help="<optional> The overflow check result output file path.",
                        required=False)


def _run_overflow_check(args):
    check_file_or_directory_path(args.input_path, True)
    OverFlowCheck(args.input_path, args.output_path).analyze()


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import Dict
from tqdm import tqdm

from msprobe.core.common.file_utils import save_json, check_path_before_create, check_path_not_exists, \
    check_file_or_directory_path
from msprobe.core.common.log import logger
from msprobe.core.config_check.ckpt_compare.megatron_loader import load_megatron_weights
from msprobe.core.config_check.ckpt_compare.metrics import METRIC_FUNC



def compare_checkpoints(ckpt_path1, ckpt_path2, output_path) -> Dict:
    """Compare weights between two checkpoints using cosine similarity and L2 distance.
    
    Args:
        ckpt_path1 (str): Path to first checkpoint directory
        ckpt_path2 (str): Path to second checkpoint directory 
        output_path (str): Path to save comparison results JSON file

    Returns:
        Dict: Dictionary containing comparison metrics for each parameter. The dictionary has the following structure:
            {
                "param_name": {
                    "cosine_similarity": float,  # Cosine similarity between parameter tensors
                    "l2_distance": float,        # L2 distance between parameter tensors  
                    "shape": List[int]           # Shape of the parameter tensors
                },
                ...
            }
    """

    # Load both checkpoints
    check_file_or_directory_path(ckpt_path1, isdir=True)
    check_file_or_directory_path(ckpt_path2, isdir=True)
    check_path_before_create(output_path)
    check_path_not_exists(output_path)
    weights1 = load_megatron_weights(ckpt_path1)
    weights2 = load_megatron_weights(ckpt_path2)
    
    # Initialize results dictionary
    results = {}
    
    # Compare weights with matching keys
    common = set(weights1) & set(weights2)
    logger.warning(f'Parameters not in ckpt2: {set(weights1) - set(weights2)}')
    logger.warning(f'Parameters not in ckpt1: {set(weights2) - set(weights1)}')
    for key in tqdm(common):
        tensor1 = weights1[key]
        tensor2 = weights2[key]
        
        results[key] = {}
        for metric, func in METRIC_FUNC.items():
            try:
                results[key][metric] = func(tensor1, tensor2)
            except Exception as e:
                results[key][metric] = 'error'
                logger.warning(f'Error when calculate {metric} for reason: {e}')

    # Write results to JSON file
    save_json(output_path, results, indent=4)
    logger.info(f"Comparison results written to {output_path}")
    return results



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import load_npy, FileChecker, FileCheckConst
from msprobe.core.common.utils import detect_framework_by_dump_json, CompareException, check_op_str_pattern_valid
from msprobe.core.common.log import logger


def read_npy_data(dir_path, file_name):
    if not file_name:
        return None

    data_path = os.path.join(dir_path, file_name)
    path_checker = FileChecker(data_path, FileCheckConst.FILE, FileCheckConst.READ_ABLE,
                               FileCheckConst.NUMPY_SUFFIX)
    data_path = path_checker.common_check()
    data_value = load_npy(data_path)
    return data_value


def check_cross_framework(bench_json_path):
    framework = detect_framework_by_dump_json(bench_json_path)
    return framework == Const.PT_FRAMEWORK


def check_name_map_dict(name_map_dict):
    if not isinstance(name_map_dict, dict):
        logger.error("'map_dict' should be a dict, please check!")
        raise CompareException(CompareException.INVALID_OBJECT_TYPE_ERROR)
    check_op_str_pattern_valid(str(name_map_dict))



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import atexit
from multiprocessing import Pool
import os
import re
import time

import numpy as np
import pandas as pd
import mindspore as ms
from mindspore import nn, ops

from msprobe.core.common.const import Const as CoreConst
from msprobe.core.common.const import FileCheckConst
from msprobe.core.common.file_utils import (
    load_npy, save_json, remove_path, load_yaml,
    create_directory, read_csv, write_df_to_csv, write_csv, move_file, move_directory)
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.dump.dump_processor.cell_dump_process import CellDumpConfig


CONSTRUCT_FILE_NAME = "construct.json"
DEFAULT_RANK_DIR = "rank0"
KEY_LAYERS = "layers"
construct = {}
cell_list = []
KEY_SIDE_EFFECT = "side_effect_io"
KEY_TOPLAYER = "TopLayer"
KEY_FORWARD = CoreConst.FORWARD
KEY_BACKWARD = CoreConst.BACKWARD
KEY_INPUT = CoreConst.INPUT
KEY_OUTPUT = CoreConst.OUTPUT
if (ms.__version__ > "2.7.0"):
    KEY_DUMP_TENSOR_DATA = "dump_tensor_data/"
else:
    KEY_DUMP_TENSOR_DATA = "dump_tensor_data_"
KEY_STATISTIC_CSV = "statistic.csv"
KEY_TD_FLAG = "td_flag"
td = ops.TensorDump()
if (ms.__version__ >= "2.5.0"):
    td_in = ops.TensorDump("in")
else:
    td_in = ops.TensorDump()
graph_step_flag = True
try:
    from mindspore._c_expression import _set_init_iter
except ImportError:
    graph_step_flag = False
td.add_prim_attr(KEY_SIDE_EFFECT, False)
td_in.add_prim_attr(KEY_SIDE_EFFECT, False)
td.add_prim_attr(KEY_TD_FLAG, True)
td_in.add_prim_attr(KEY_TD_FLAG, True)
dump_task = CoreConst.STATISTICS
np_ms_dtype_dict = {
    "bool": ms.bool_,
    "int8": ms.int8,
    "byte": ms.byte,
    "int16": ms.int16,
    "short": ms.short,
    "int32": ms.int32,
    "intc": ms.intc,
    "int64": ms.int64,
    "intp": ms.intp,
    "uint8": ms.uint8,
    "ubyte": ms.ubyte,
    "uint16": ms.uint16,
    "ushort": ms.ushort,
    "uint32": ms.uint32,
    "uintc": ms.uintc,
    "uint64": ms.uint64,
    "uintp": ms.uintp,
    "float16": ms.float16,
    "half": ms.half,
    "float32": ms.float32,
    "single": ms.single,
    "float64": ms.float64,
    "double": ms.double,
    "bfloat16": ms.bfloat16,
    "complex64": ms.complex64,
    "complex128": ms.complex128
}


def gen_file_path(dump_path, cell_prefix, suffix, io_type, index):
    data_path = os.path.join(dump_path, '{step}', '{rank}', CoreConst.DUMP_TENSOR_DATA)
    file_name = ""
    if dump_task == CoreConst.TENSOR:
        file_name = cell_prefix + CoreConst.SEP + suffix + CoreConst.SEP + io_type + CoreConst.SEP + str(index)
    if dump_task == CoreConst.STATISTICS:
        file_name = cell_prefix + CoreConst.HYPHEN + suffix + CoreConst.HYPHEN + io_type + CoreConst.HYPHEN + str(index)
    return os.path.join(data_path, file_name)


def partial_func(func, dump_path, cell_prefix, index, io_type):
    def newfunc(*args, **kwargs):
        return func(dump_path, cell_prefix, index, io_type, *args, **kwargs)
    return newfunc


def clip_gradient(dump_path, cell_prefix, index, io_type, dx):
    if io_type == KEY_OUTPUT:
        temp = td(gen_file_path(dump_path, cell_prefix, KEY_BACKWARD, io_type, index), dx)
        dx = ops.depend(dx, temp)
    elif io_type == KEY_INPUT:
        temp = td_in(gen_file_path(dump_path, cell_prefix, KEY_BACKWARD, io_type, index), dx)
        dx = ops.depend(dx, temp)
    return dx


def need_tensordump_in(cell_obj, attr):
    return hasattr(cell_obj, attr) and getattr(cell_obj, attr) == "in"


def cell_construct_wrapper(func, self):
    def new_construct(self, *args, **kwargs):
        new_args = []
        out_list = []

        index = 0
        item = None
        backward_or_all = self.data_mode in ["backward", "all"]
        forward_or_all = self.data_mode in ["forward", "all"]
        # The inputs of the cell.
        for index, item in enumerate(args):
            if backward_or_all and ops.is_tensor(item):
                item = self.output_clips[index](item)
            if forward_or_all and ops.is_tensor(item):
                if need_tensordump_in(self, 'input_dump_mode'):
                    temp = td_in(
                        gen_file_path(self.dump_path, self.cell_prefix, KEY_FORWARD, KEY_INPUT, index),
                        item
                    )
                else:
                    temp = td(
                        gen_file_path(self.dump_path, self.cell_prefix, KEY_FORWARD, KEY_INPUT, index),
                        item
                    )
                item = ops.depend(item, temp)
            new_args.append(item)

        out = func(*new_args, **kwargs)

        # The outputs of the cell.
        if isinstance(out, tuple):
            for index, item in enumerate(out):
                if backward_or_all and ops.is_tensor(item):
                    item = self.input_clips[index](item)
                if forward_or_all and ops.is_tensor(item):
                    if need_tensordump_in(self, 'output_dump_mode'):
                        temp = td_in(
                            gen_file_path(self.dump_path, self.cell_prefix, KEY_FORWARD, KEY_OUTPUT, index),
                            item
                        )
                    else:
                        temp = td(
                            gen_file_path(self.dump_path, self.cell_prefix, KEY_FORWARD, KEY_OUTPUT, index),
                            item
                        )
                    item = ops.depend(item, temp)
                    out_list.append(item)
                elif forward_or_all and not ops.is_tensor(item):
                    out_list.append(item)
            out_list = tuple(out_list)
            return out_list
        else:
            if backward_or_all:
                out = self.input_clips[0](out)
            if forward_or_all and ops.is_tensor(out):
                if need_tensordump_in(self, 'output_dump_mode'):
                    temp = td_in(
                        gen_file_path(self.dump_path, self.cell_prefix, KEY_FORWARD, KEY_OUTPUT, 0),
                        out
                    )
                else:
                    temp = td(
                        gen_file_path(self.dump_path, self.cell_prefix, KEY_FORWARD, KEY_OUTPUT, 0),
                        out
                    )
                out = ops.depend(out, temp)
            return out

    return new_construct.__get__(self, type(self))


# 获取目录下所有文件名并根据TensorDump落盘自增id从小到大排序
def sort_filenames(path):
    filenames = os.listdir(path)
    id_pattern = re.compile(rf'{CoreConst.REPLACEMENT_CHARACTER}(\d+){CoreConst.NUMPY_SUFFIX}$')
    # 只保留能提取到数字id的文件，避免数组越界
    valid_files = []
    for filename in filenames:
        match = id_pattern.findall(filename)
        if match and match[0].isdigit():
            valid_files.append(filename)
        else:
            logger.warning(f"File {filename} does not match the expected pattern and will be ignored.")
    valid_files.sort(key=lambda x: int(id_pattern.findall(x)[0]))
    return valid_files


def rename_filename(path="", data_df=None):
    if dump_task == CoreConst.TENSOR:
        filenames = sort_filenames(path)
    if dump_task == CoreConst.STATISTICS:
        filenames = data_df[CoreConst.OP_NAME].tolist()

    filename_dict = {}
    for index, filename in enumerate(filenames):
        if dump_task == CoreConst.TENSOR:
            name_field = filename.rsplit(CoreConst.REPLACEMENT_CHARACTER, 1)[0]
        if dump_task == CoreConst.STATISTICS:
            name_field = filename

        if name_field in filename_dict:
            filename_dict[name_field] += 1
        else:
            filename_dict[name_field] = 0

        cell_index = filename_dict[name_field]

        # 修改文件名，增加重复调用Cell的序号
        if CoreConst.FORWARD_PATTERN in filename:
            # Format: Cell.{cell_name}.{class_name}.{forward/backward}.{number}.{input/output}.{index}_{dtype}_{id}.npy
            new_file_name = filename.replace(CoreConst.FORWARD_PATTERN,
                                             CoreConst.FORWARD_PATTERN + str(cell_index) + CoreConst.SEP)
        if CoreConst.BACKWARD_PATTERN in filename:
            new_file_name = filename.replace(CoreConst.BACKWARD_PATTERN,
                                             CoreConst.BACKWARD_PATTERN + str(cell_index) + CoreConst.SEP)
        if dump_task == CoreConst.TENSOR:
            move_file(os.path.join(path, filename), os.path.join(path, new_file_name))
        if dump_task == CoreConst.STATISTICS:
            data_df.loc[index, CoreConst.OP_NAME] = new_file_name
    logger.info("==========The rename_filename phase is Finished!==========")


# Extract the field between the first "." and the third to last ".", i.e. {cell_name}
def get_cell_name(string):
    parts = string.split(CoreConst.SEP)
    if len(parts) < 4:
        return None
    start_index = 1
    end_index = len(parts) - 3
    return CoreConst.SEP.join(parts[start_index:end_index])


# Extract the field between the last "." and the second to last ".", i.e. {data_made}
def get_data_mode(string):
    last_dot_index = string.rfind(CoreConst.SEP)
    second_last_dot_index = string.rfind(CoreConst.SEP, 0, last_dot_index)
    data_mode = string[second_last_dot_index + 1:last_dot_index]
    return data_mode


# 判断二者之间是否存在父子关系
def check_relation(cell_name, parent_cell_name):
    layers_pattern = rf"{CoreConst.SEP}{KEY_LAYERS}{CoreConst.SEP}\d+$"
    last_dot_index = cell_name.rfind(CoreConst.SEP)
    if last_dot_index == -1:
        return False
    # 如果cell_name最后一个'.'之前的字段等于parent_cell_name，则判定存在父子关系
    sub_cell_name = cell_name[:last_dot_index]
    if sub_cell_name == parent_cell_name:
        return True
    elif re.search(layers_pattern, cell_name):
        # 如果cell_name以".layer.{layer_id}"结尾，且去掉该字段后等于parent_cell_name，则判定存在父子关系
        sub_cell_name = re.sub(layers_pattern, '', cell_name)
        if sub_cell_name == parent_cell_name:
            return True
    return False


def get_construct(cell_list_input):
    for cell in cell_list_input:
        cell_name = get_cell_name(cell)
        cell_data_mode = get_data_mode(cell)
        found_flag = False
        for parent_cell in cell_list_input:
            parent_cell_name = get_cell_name(parent_cell)
            parent_data_mode = get_data_mode(parent_cell)
            has_relation = check_relation(cell_name, parent_cell_name)
            if has_relation and parent_data_mode == cell_data_mode:
                construct.update({cell: parent_cell})
                found_flag = True
                break
        if not found_flag:
            construct.update({cell: None})


def generate_construct(path):
    global construct
    if dump_task == CoreConst.TENSOR:
        # filename格式：Cell.clip_grad_norm.ClipGradNorm.forward.0.output.1_int32_0.npy
        filenames = sort_filenames(path)
        point_position = 3
    if dump_task == CoreConst.STATISTICS:
        df = read_csv(path)
        # filename格式：Cell.clip_grad_norm.ClipGradNorm.forward.0.output.1
        filenames = df[CoreConst.OP_NAME].tolist()
        point_position = 2

    # 提取文件名中Cell.{cell_name}.{class_name}.{data_mode}.{重复调用此cell的序号}字段，并存入cell_list
    for filename in filenames:
        mid_field = filename.rsplit(CoreConst.SEP, point_position)[0]
        if KEY_INPUT in filename:
            if mid_field in cell_list:
                cell_list.remove(mid_field)
            cell_list.append(mid_field)
        else:
            if mid_field not in cell_list:
                index = filenames.index(filename)
                output_field = mid_field + KEY_OUTPUT
                find_flag = False
                for filename_other in cell_list[index + 1:]:
                    if output_field in filename_other:
                        find_flag = True
                if find_flag is False:
                    cell_list.append(mid_field)

    get_construct(cell_list)

    # 生成JSON文件
    rank_dir = os.path.dirname(path)
    json_path = os.path.join(rank_dir, CONSTRUCT_FILE_NAME)
    save_json(json_path, construct, indent=1)

    # 清空'construct'继续处理下一个路径下的数据
    construct = {}
    logger.info(f"Construct data saved to {json_path}")


def process_file(file_path):
    try:
        # 读取.npy文件内容
        npy_content = load_npy(file_path)
        logger.debug(f"Loaded {file_path}: shape is {npy_content.shape}, dtype is {npy_content.dtype}")

        # 文件名举例:Cell.network._backbone.loss.CrossEntropyLoss.forward.0.input.0_float32_165.npy
        parts = os.path.basename(file_path).split(CoreConst.SEP)
        data_dtype = ""
        # 获取0_float32_165或者0_in_float32_165中的float32
        data_dtype_list = parts[-2].split('_')
        if len(data_dtype_list) > 1:
            data_dtype = data_dtype_list[-2]
        # op_name是Cell.network._backbone.loss.CrossEntropyLoss.forward.0
        op_name = CoreConst.SEP.join(parts[:-3])
        ms_dtype = np_ms_dtype_dict.get(data_dtype)
        if ms_dtype is None:
            logger.warning(f"Get dtype None from file {file_path}")

        # 修改落盘文件名字，去掉TensorDump自带的数据类型和自增id字段
        data_file_name = os.path.basename(file_path)
        data_file_dir = os.path.dirname(file_path)
        parts = data_file_name.split(CoreConst.SEP)
        if len(parts) >= 2:
            param_index = parts[-2].split(CoreConst.REPLACEMENT_CHARACTER)[0]
            pre_parts = CoreConst.SEP.join(parts[:-2])
            new_file_name = pre_parts + CoreConst.SEP + param_index + CoreConst.NUMPY_SUFFIX
            move_file(os.path.join(data_file_dir, data_file_name), os.path.join(data_file_dir, new_file_name))
            logger.debug(f"{data_file_name} is renamed to {new_file_name}")
        else:
            logger.warning(f"Failed to rename {data_file_name}.")
            new_file_name = data_file_name

        tensor_json = {
            CoreConst.TYPE: 'mindspore.Tensor',
            CoreConst.DTYPE: str(ms_dtype),
            CoreConst.SHAPE: list(npy_content.shape),
            CoreConst.MAX: npy_content.max().item(),
            CoreConst.MIN: npy_content.min().item(),
            CoreConst.MEAN: npy_content.mean().item(),
            CoreConst.NORM: np.linalg.norm(npy_content).item(),
            CoreConst.DATA_NAME: new_file_name
        }

        # 根据文件名的最后一个部分（输入或输出）确定是添加到input_args还是output
        if parts[-3] == KEY_INPUT:
            return op_name, CoreConst.INPUT_ARGS, tensor_json
        elif parts[-3] == KEY_OUTPUT:
            return op_name, KEY_OUTPUT, tensor_json
        else:
            return None, None, None

    except Exception as e:
        logger.error(f"Error reading {file_path}: {e}")
        return None, None, None


def custom_sort(item, key_to_index):
    key = item[0]
    return key_to_index.get(key, float('inf'))


def convert_special_values(value):
    if isinstance(value, str):
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        try:
            return float(value)
        except ValueError:
            return value
    elif pd.isna(value):
        return None
    return value


def process_csv(path):
    data_info = []
    df = read_csv(path)
    df = df.sort_values(by='Op Name', ascending=True)
    columns = df.columns
    colume_to_json_key = {
        'Max Value': CoreConst.MAX,
        'Min Value': CoreConst.MIN,
        'Avg Value': CoreConst.MEAN,
        'L2Norm Value': CoreConst.NORM
    }
    for _, row in df.iterrows():
        # op_name_value格式：Cell.network._backbone.loss.CrossEntropyLoss.forward.0.input.0
        op_name_value = row['Op Name']
        op_name = op_name_value.rsplit(CoreConst.SEP, 2)[0]

        # 获取input/output字段
        io_key = op_name_value.split(CoreConst.SEP)[-2]

        # shape读取出来为字符串类型转为list。"(1,4096)"->[1,4096]
        shape_num = re.findall(r'\d+', row['Shape'])
        shape = [int(num) for num in shape_num]

        tensor_json = {
            CoreConst.TYPE: 'mindspore.Tensor',
            CoreConst.DTYPE: str(np_ms_dtype_dict.get(row['Data Type'])),
            CoreConst.SHAPE: shape
        }
        for col_name, json_key in colume_to_json_key.items():
            if col_name in columns:
                value = convert_special_values(row[col_name])
                tensor_json[json_key] = value
      
        if io_key == KEY_INPUT:
            data_info.append([op_name, CoreConst.INPUT_ARGS, tensor_json])
        elif io_key == KEY_OUTPUT:
            data_info.append([op_name, KEY_OUTPUT, tensor_json])
        else:
            data_info.append([None, None, None])
    return data_info


def generate_dump_info(path):
    if not os.path.exists(path):
        logger.error("The provided path does not exist.")
        return

    if dump_task == CoreConst.TENSOR:
        dump_data = {"task": "tensor", "level": "L0", "framework": "mindspore", "dump_data_dir": path, "data": {}}
        with Pool(processes=10) as pool:
            file_paths = []
            for file in os.listdir(path):
                if file.endswith(FileCheckConst.NUMPY_SUFFIX):
                    file_paths.append((os.path.join(path, file),))
            file_paths.sort()
            results = pool.starmap(process_file, file_paths)
    if dump_task == CoreConst.STATISTICS:
        dump_data = {"task": "statistics", "level": "L0", "framework": "mindspore", "dump_data_dir": None, "data": {}}
        results = process_csv(path)

    # 收集结果
    for op_name, key, tensor_json in results:
        if op_name:
            if op_name not in dump_data.get(CoreConst.DATA, {}):
                dump_data.get(CoreConst.DATA, {})[op_name] = {CoreConst.INPUT_ARGS: [],
                                                              CoreConst.INPUT_KWARGS: {},
                                                              KEY_OUTPUT: []}
            if key not in dump_data.get(CoreConst.DATA, {}).get(op_name, {}):
                dump_data.get(CoreConst.DATA, {}).get(op_name, {})[key] = []
            dump_data.get(CoreConst.DATA, {}).get(op_name, {}).get(key, []).append(tensor_json)

    # 根据cell_list排序
    data_dict = dump_data.get(CoreConst.DATA, {})
    key_to_index = {key: index for index, key in enumerate(cell_list)}
    sorted_data_dict = dict(sorted(data_dict.items(), key=lambda item: custom_sort(item, key_to_index)))
    dump_data[CoreConst.DATA] = sorted_data_dict

    # 将数据写入dump.json
    json_path = os.path.join(os.path.dirname(path), 'dump.json')
    save_json(json_path, dump_data, indent=1)

    logger.info(f"Dump data saved to {json_path}")


def generate_stack_info(path):
    if not os.path.exists(path):
        logger.error("The provided path does not exist.")
        return

    stack_data = {}
    for cell_name in cell_list:
        stack_data.update({cell_name: []})

    # 将数据写入stack.json
    json_path = os.path.join(os.path.dirname(path), 'stack.json')
    save_json(json_path, stack_data, indent=1)

    # 删除csv文件
    if dump_task == CoreConst.STATISTICS:
        remove_path(path)

    logger.info(f"Stack data saved to {json_path}")


def is_download_finished(directory, interval=3):
    """
    判断指定目录在一段时间后是否有数据被下载完成
    :param directory: 指定目录的路径
    :param interval: 检查的时间间隔（秒），默认为 3 秒
    :return: 如有数据被下载完成返回 True，否则返回 False
    """
    # 检查目录是否存在
    if not os.path.exists(directory):
        logger.warning(f"The specified directory {directory} does not exist.")
        return False, False
    initial_modification_time = os.path.getmtime(directory)
    time.sleep(interval)
    current_modification_time = os.path.getmtime(directory)
    # 比较初始和当前修改时间
    if current_modification_time > initial_modification_time:
        return False, True
    else:
        return True, False


def process(dump_path):
    rank_id = os.environ.get('RANK_ID')
    rank_dir = DEFAULT_RANK_DIR
    if rank_id is not None:
        rank_dir = CoreConst.RANK + str(rank_id)

    step_dir_list = os.listdir(dump_path)
    for step_dir in step_dir_list:
        step_path = os.path.join(dump_path, step_dir)
        rank_path = os.path.join(step_path, rank_dir)
        npy_path = os.path.join(rank_path, CoreConst.DUMP_TENSOR_DATA)
        check_times = 0
        while True:
            is_finished, is_downloading = is_download_finished(npy_path)
            if not is_finished:
                if not is_downloading:
                    logger.warning(f'{npy_path} does not exist.')
                    break
                check_times += 1
                if check_times < 1000:
                    logger.info("There is data being downloaded in the specified directory, continue checking...")
                else:
                    logger.warning('Download timeout, stop checking.')
                    break
            else:
                logger.info("There is no data being downloaded in the specified directory, stop checking.")
                break
        logger.info("==========Start processing data that has already been stored on the disk!==========")
        rename_filename(path=npy_path)
        generate_construct(npy_path)
        generate_dump_info(npy_path)
        generate_stack_info(npy_path)
        # 单卡场景，rank目录名称为rank
        if rank_id is None:
            new_rank_path = os.path.join(step_path, CoreConst.RANK)
            try:
                move_directory(rank_path, new_rank_path)
                logger.debug(f"Directory was successfully renamed to: {new_rank_path}")
            except Exception as e:
                logger.warning(f"Failed to renamed to {new_rank_path}: {e}")
        logger.info("==========JSON file generation completed!==========")


# 删除csv文件中每行数据最后面的逗号
def remove_trailing_commas(filename):
    csv_data = read_csv(filename, as_pd=False)
    for i in range(1, len(csv_data)):
        if csv_data[i] and csv_data[i][-1] == "":
            csv_data[i].pop()
    write_csv(csv_data, filename, mode="w")


# 将相同step的csv文件合并，并加工后存入相应step目录下
def merge_file(dump_path, rank_dir, file_dict):
    rank_dir = rank_dir.replace(CoreConst.REPLACEMENT_CHARACTER, '')
    for step_dir, file_list in file_dict.items():
        step_dir = CoreConst.STEP + step_dir
        rank_path = os.path.join(dump_path, step_dir, rank_dir)
        create_directory(rank_path)
        output_file = os.path.join(rank_path, KEY_STATISTIC_CSV)

        all_dfs = []
        try:
            for file_path in file_list:
                remove_trailing_commas(file_path)
                df = read_csv(file_path)
                all_dfs.append(df)

            # 合并所有 DataFrame
            merged_df = pd.concat(all_dfs, ignore_index=True)
          
            # 按 Timestamp 字段升序排序
            merged_df = merged_df.sort_values(by='Timestamp', ascending=True)
            # 删除Slot字段为0的数据
            merged_df = merged_df[merged_df['Slot'] != 0]
            # 重置索引，从0开始排序
            merged_df.reset_index(drop=True, inplace=True)

            # 获取op_name并加工为Cell.network._backbone.LlamaForCausalLM.forward.input.0格式
            merged_df[CoreConst.OP_NAME] = merged_df[CoreConst.OP_NAME].str.split(KEY_DUMP_TENSOR_DATA, expand=True)[1]
            merged_df[CoreConst.OP_NAME] = (
                merged_df[CoreConst.OP_NAME].str.split(CoreConst.PIPE_SEPARATOR, expand=True)[0]
            )
            merged_df[CoreConst.OP_NAME] = (
                merged_df[CoreConst.OP_NAME].str.replace(CoreConst.HYPHEN, CoreConst.SEP, regex=False)
            )
            # 重命名op_name，改为Cell.{cell_name}.{class_name}.{forward/backward}.{number}.{input/output}.{index}格式
            rename_filename(data_df=merged_df)

            # 将合并并排序后的 DataFrame 保存到相应step目录下
            write_df_to_csv(merged_df, output_file)
        except FileNotFoundError:
            logger.error("One or more files not found.")
        except KeyError:
            logger.error("The value of the ‘Op Name’ field does not contain KEY_DUMP_TENSOR_DATA,"
                         " and the index is out of bounds.")
        except Exception as e:
            logger.error(f"An error occurred:{e}")


def process_statistics(dump_path):
    rank_id = os.environ.get('RANK_ID')
    rank_dir_kbk = "rank_0"
    if rank_id is not None:
        rank_dir_kbk = CoreConst.RANK + CoreConst.REPLACEMENT_CHARACTER + str(rank_id)
    rank_path_kbk = os.path.join(dump_path, rank_dir_kbk)

    # 按相同step数将csv文件名分组存入file_dict
    file_dict = {}
    depth_limit = 4
    base_depth = rank_path_kbk.count(os.sep)
    for root, _, files in os.walk(rank_path_kbk):
        current_depth = root.count(os.sep) - base_depth
        if current_depth > depth_limit:
            continue
        for file in files:
            if file == KEY_STATISTIC_CSV:
                file_path = os.path.join(root, file)
                step_dir = os.path.basename(os.path.dirname(file_path))
                if step_dir in file_dict:
                    file_dict[step_dir].append(file_path)
                else:
                    file_dict[step_dir] = [file_path]

    # 将相同step的csv文件合并，并加工后存入相应step目录下
    merge_file(dump_path, rank_dir_kbk, file_dict)

    rank_dir = rank_dir_kbk.replace(CoreConst.REPLACEMENT_CHARACTER, '')
    dir_list = os.listdir(dump_path)
    step_dir_list = [d for d in dir_list if d.startswith(CoreConst.STEP)]
    for step_dir in step_dir_list:
        step_path = os.path.join(dump_path, step_dir)
        rank_path = os.path.join(step_path, rank_dir)
        csv_path = os.path.join(rank_path, KEY_STATISTIC_CSV)
        logger.info("==========Start processing data csv!==========")
        generate_construct(csv_path)
        generate_dump_info(csv_path)
        generate_stack_info(csv_path)
        remove_path(rank_path_kbk)
        # 单卡场景，rank目录名称为rank
        if rank_id is None:
            new_rank_path = os.path.join(step_path, CoreConst.RANK)
            try:
                move_directory(rank_path, new_rank_path)
                logger.info(f"Directory was successfully renamed to: {new_rank_path}")
            except Exception as e:
                logger.warning(f"Failed to renamed to {new_rank_path}: {e}")
        logger.info("==========JSON file generation completed!==========")


def get_yaml_keys(yaml_data):
    keys = []
    for key, _ in yaml_data.items():
        keys.append(key)
    return keys


def get_tensordump_mode(input_str):
    left_index = input_str.find('(')
    right_index = input_str.find(')')

    # 提取括号内的字符串
    if left_index != -1 and right_index != -1:
        inner_str = input_str[left_index + 1:right_index]
        # 分割字符串得到元素列表
        elements = inner_str.split(',')
        if len(elements) >= 2:
            # 去除元素前后的空格
            first_element = elements[0].strip()
            second_element = elements[1].strip()
            return first_element, second_element
    return None, None


def set_tensordump_mode(cell, input_str):
    first_str, second_str = get_tensordump_mode(input_str)
    if first_str and second_str:
        cell.input_dump_mode = first_str
        cell.output_dump_mode = second_str


def create_kbyk_json(dump_path, summary_mode, step):
    if step:
        step_str = ""
        for s in step:
            step_str += (str(s) + '|')
        iteration = step_str[:-1]
    else:
        iteration = "all"

    if summary_mode == "statistics":
        statistic_category = ["max", "min", "avg", "l2norm"]
    elif "mean" in summary_mode:
        mean_index = summary_mode.index("mean")
        summary_mode[mean_index] = "avg"
        statistic_category = summary_mode
    else:
        statistic_category = summary_mode

    config_json = {
        "common_dump_settings": {
            "op_debug_mode": 0,
            "dump_mode": 1,
            "path": dump_path,
            "net_name": "Net",
            "iteration": iteration,
            "saved_data": "statistic",
            "input_output": 0,
            "kernels": ["TensorDump"],
            "support_device": [0, 1, 2, 3, 4, 5, 6, 7],
            "statistic_category": statistic_category
        },
        "e2e_dump_settings": {
            "enable": False,
            "trans_flag": True,
            "stat_calc_mode": "device"
        }
    }

    create_directory(dump_path)
    rank_id = os.environ.get('RANK_ID')
    if rank_id is None:
        rank_id = 0
    config_json_path = os.path.join(dump_path, str(rank_id) + "kernel_kbyk_dump.json")
    save_json(config_json_path, config_json, indent=4)
    logger.info(config_json_path + " has been created.")
    return config_json_path


def start(config: CellDumpConfig):
    global dump_task
    dump_task = config.task
    net = config.net
    dump_path = config.dump_path
    data_mode = config.data_mode
    summary_mode = config.summary_mode
    step = config.step
    if dump_task == CoreConst.STATISTICS:
        # 使能KBK dump
        config_json_path = create_kbyk_json(dump_path, summary_mode, step)
        os.environ["MINDSPORE_DUMP_CONFIG"] = config_json_path

        # 执行过程中跳过TensorDump算子
        os.environ["MS_KERNEL_LAUNCH_SKIP"] = "TensorDump"

        # 初始化静态图KBK dump的step数，从0开始
        if not graph_step_flag:
            raise Exception(
                "Importing _set_init_iter failed, "
                "please use the latest version package of MindSpore."
            )
        _set_init_iter(0)
        remove_path(config_json_path)

    if net is None:
        return

    if isinstance(net, nn.Cell):
        net = (('', net),)

    td_config_path = ""
    try:
        import mindformers
        mindformers_file = mindformers.__file__
        mindformers_dir = os.path.dirname(mindformers_file)
        td_config_path = os.path.join(mindformers_dir, "configuration", "layer_mapping.yaml")
        if not os.path.exists(td_config_path):
            td_config_path = ""
            logger.warning("The configuration file in mindformers was not loaded, the default mode will be used.")
    except ImportError:
        logger.warning("The mindFormers failed to load, the default mode will be used.")

    if td_config_path == "":
        yaml_data = {}
    else:
        yaml_data = load_yaml(td_config_path)
    first_layer_key = get_yaml_keys(yaml_data)

    black_list = ["grad_reducer", ""]

    for name_and_model in net:
        for name, cell in name_and_model[1].cells_and_names(name_prefix=name_and_model[0]):
            class_name = cell.__class__.__name__
            # 跳过黑名单cell
            if name in black_list:
                logger.info(f"Cell {name}.{class_name} is skipped!")
                continue
            # 跳过框架内部的cell
            if class_name.startswith(CoreConst.REPLACEMENT_CHARACTER):
                logger.info(f"Cell {name}.{class_name} is skipped!")
                continue
            else:
                # Format: Cell.{cell_name}.{class_name}
                cell.cell_prefix = CoreConst.SEP.join([CoreConst.CELL, name, cell.__class__.__name__])
                if dump_task == CoreConst.STATISTICS:
                    cell.cell_prefix = cell.cell_prefix.replace(CoreConst.SEP, CoreConst.HYPHEN)

            # 根据yaml配置文件设置cell的TensorDump模式
            if class_name in first_layer_key:
                layer_data = yaml_data.get(class_name)
                if layer_data:
                    for child_name, child_cell in cell.cells_and_names():
                        if child_name in layer_data:
                            set_tensordump_mode(child_cell, layer_data[child_name])
            top_layer_data = yaml_data.get(KEY_TOPLAYER)
            if top_layer_data and name in top_layer_data:
                set_tensordump_mode(cell, top_layer_data[name])

            # 替换construct函数
            cell.construct = cell_construct_wrapper(cell.construct, cell)
            logger.info(f"Cell {name}: construct function is wrapped!")
            cell.dump_path = dump_path
            cell.data_mode = data_mode
            cell.input_clips = []
            cell.output_clips = []
            # It is assumed that each cell has a maximum of 50 outputs and 50 inputs.
            for i in range(50):
                cell.input_clips.append(
                    ops.InsertGradientOf(partial_func(clip_gradient, cell.dump_path, cell.cell_prefix, i, KEY_INPUT))
                )
                cell.output_clips.append(
                    ops.InsertGradientOf(partial_func(clip_gradient, cell.dump_path, cell.cell_prefix, i, KEY_OUTPUT))
                )

    logger.info("==========The cell_dump_process_start phase is Finished!==========")
    if dump_task == CoreConst.TENSOR:
        atexit.register(process, dump_path=dump_path)
    if dump_task == CoreConst.STATISTICS:
        atexit.register(process_statistics, dump_path=dump_path)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.config_check.config_checker import ConfigChecker
from msprobe.core.config_check.ckpt_compare.ckpt_comparator import compare_checkpoints
from msprobe.core.common.log import logger


def pack(shell_path, output_path):
    ConfigChecker(shell_path=shell_path, output_zip_path=output_path)


def compare(bench_zip_path, cmp_zip_path, output_path):
    ConfigChecker.compare(bench_zip_path, cmp_zip_path, output_path)


def _config_checking_parser(parser):
    parser.add_argument('-d', '--dump', nargs='*', help='Collect the train config into a zip file')
    parser.add_argument('-c', '--compare', nargs=2, help='Compare two zip files or checkpoints')
    parser.add_argument('-o', '--output', help='output path, default is ./config_check_result')


def _run_config_checking_command(args):
    if args.dump is not None:
        output_dirpath = args.output if args.output else "./config_check_pack.zip"
        pack(args.dump, output_dirpath)
    elif args.compare:
        if args.compare[0].endswith('zip'):
            logger.info('The input paths is zip files, comparing packed config.')
            output_dirpath = args.output if args.output else "./config_check_result"
            compare(args.compare[0], args.compare[1], output_dirpath)
        else:
            logger.info('Comparing model checkpoint.')
            output_dirpath = args.output if args.output else "./ckpt_similarity.json"
            compare_checkpoints(args.compare[0], args.compare[1], output_dirpath)

    else:
        logger.error("The param is not correct, you need to give '-d' for dump or '-c' for compare.")
        raise Exception("The param is not correct, you need to give '-d' for dump or '-c' for compare.")



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import create_directory, save_json
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.dump.debugger.debugger_config import DebuggerConfig

import mindspore as ms
ms_version = ms.__version__


class KernelKbykDump:
    COMMON_SETTINGS = "common_dump_settings"
    E2E_SETTINGS = "e2e_dump_settings"

    def __init__(self, config: DebuggerConfig):
        self.dump_json = dict()
        common_set = dict()
        e2e_set = dict()

        common_set = dict()
        common_set["dump_mode"] = 0
        common_set["path"] = ""
        common_set["net_name"] = "Net"
        common_set["iteration"] = "all"
        common_set["saved_data"] = "statistic"
        common_set["input_output"] = 0
        common_set["kernels"] = []
        common_set["support_device"] = [0, 1, 2, 3, 4, 5, 6, 7]
        common_set["statistic_category"] = []

        if config.stat_cal_mode and config.device_stat_precision_mode:
            e2e_set = {
                        "enable": not config.async_dump,
                        "trans_flag": True,
                        "stat_calc_mode": config.stat_cal_mode,
                        "device_stat_precision_mode": config.device_stat_precision_mode
                    }
        else:
            e2e_set = {
                        "enable": not config.async_dump,
                        "trans_flag": True
                    }

        if config.list:
            common_set["dump_mode"] = 1
            common_set["kernels"] = config.list
        common_set["path"] = config.dump_path
        if config.step:
            step_str = ""
            for s in config.step:
                step_str += (str(s) + '|')
            common_set["iteration"] = step_str[:-1]
        if config.rank:
            common_set["support_device"] = config.rank
        if config.task == Const.TENSOR:
            common_set["saved_data"] = Const.TENSOR
        if len(config.data_mode) == 1:
            if config.data_mode[0] == Const.INPUT:
                common_set["input_output"] = 1
            if config.data_mode[0] == Const.OUTPUT:
                common_set["input_output"] = 2
        if config.summary_mode:
            if isinstance(config.summary_mode, str):
                if config.summary_mode == Const.STATISTICS:
                    common_set["statistic_category"] = ["max", "min", "avg", "l2norm"]
                else:
                    mode = self._process_hash(config.summary_mode)
                    common_set["statistic_category"] = [mode]
            elif isinstance(config.summary_mode, list):
                common_set["statistic_category"] = list({
                    self._process_hash("avg" if mode == "mean" else mode)
                    for mode in config.summary_mode
                })

        self.dump_json[KernelKbykDump.COMMON_SETTINGS] = common_set
        self.dump_json[KernelKbykDump.E2E_SETTINGS] = e2e_set

    @staticmethod
    def _process_hash(value):
        if ms_version <= "2.7.0" and (value == Const.HASH or value == Const.MD5):
            value = "md5"
        elif value == Const.MD5:
            value = "hash:md5"
        return value

    def handle(self):
        json_path = self.dump_json[KernelKbykDump.COMMON_SETTINGS]["path"]
        create_directory(json_path)
        json_path = os.path.join(json_path, "kernel_kbyk_dump.json")
        save_json(json_path, self.dump_json, indent=4)
        logger.info(json_path + " has been created.")

        os.environ["MINDSPORE_DUMP_CONFIG"] = json_path
        if "MS_ACL_DUMP_CFG_PATH" in os.environ:
            del os.environ["MS_ACL_DUMP_CFG_PATH"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

# 定义比对算法及比对标准
import math
import torch
import numpy as np

from msprobe.pytorch.api_accuracy_checker.compare.compare_utils import ULP_PARAMETERS
from msprobe.pytorch.api_accuracy_checker.precision_standard.standard_config import StandardConfig
from msprobe.pytorch.api_accuracy_checker.common.utils import is_dtype_fp8
from msprobe.core.common.const import CompareConst


#cos
def cosine_sim(bench_output, device_output):
    msg = ""
    n_value = device_output.reshape(-1)
    b_value = bench_output.reshape(-1)
    cos = CompareConst.SPACE
    np.seterr(divide="ignore", invalid="ignore")
    if n_value.shape != b_value.shape:
        msg = f"Shape of device and bench outputs don't match. device: {n_value.shape}, bench: {b_value.shape}."
        return -1, False, msg
    if len(n_value) == 1:
        msg = "All the data in device dump data is scalar. Please refer to other compare algorithms."
        return cos, True, msg
    n_value_max = np.max(np.abs(n_value))
    b_value_max = np.max(np.abs(b_value))
    if n_value_max <= np.finfo(float).eps and b_value_max <= np.finfo(float).eps:
        msg = "All the data in device and bench outputs are zero."
        return cos, True, msg
    elif n_value_max <= np.finfo(float).eps:
        msg = "All the data is zero in device dump data."
        return CompareConst.SPACE, False, msg
    elif b_value_max <= np.finfo(float).eps:
        msg = "All the data is zero in bench dump data."
        return CompareConst.SPACE, False, msg
    else:
        n_value = n_value.astype(float) / n_value_max
        b_value = b_value.astype(float) / b_value_max
        cos = np.dot(n_value, b_value) / (np.linalg.norm(n_value) * np.linalg.norm(b_value))
        if np.isnan(cos):
            msg = "Dump data has NaN when comparing with Cosine Similarity."
        cos = np.clip(cos, -1, 1)
        return cos, cos > 0.99, msg


#rmse
def get_rmse(abs_err, inf_nan_mask):
    masked_ae = np.where(inf_nan_mask, 0, abs_err)
    mse = np.mean(np.square(masked_ae))
    inf_nan_cnt = np.sum(inf_nan_mask)
    mse = mse * (abs_err.size / (abs_err.size - inf_nan_cnt + 0.0001) + 0.0001)
    rmse = np.sqrt(mse)
    return rmse


#误差均衡性
def get_error_balance(bench_data, device_data):
    larger_count = np.sum(np.greater(device_data - bench_data.astype(device_data.dtype), 0))
    smaller_count = np.sum(np.less(device_data - bench_data.astype(device_data.dtype), 0))
    total_count = bench_data.size
    error_balance = abs(larger_count - smaller_count) / total_count if total_count > 0 else 0
    return error_balance


#小值域错误占比
def get_small_value_err_ratio(small_value_mask, abs_err_greater_mask):
    err_mask = np.logical_and(small_value_mask, abs_err_greater_mask)
    small_value_err_num = np.sum(err_mask)
    small_value_num = np.sum(small_value_mask)
    return 0 if small_value_num == 0 else small_value_err_num / small_value_num


def get_rel_err(abs_err, abs_bench_with_eps, small_value_mask, inf_nan_mask):
    rel_err_tmp = abs_err / abs_bench_with_eps
    rel_err_mask = np.logical_or(small_value_mask, inf_nan_mask)
    rel_err = np.where(rel_err_mask, -1, rel_err_tmp)
    return rel_err


def get_abs_err(bench_data, device_data):
    abs_err = np.abs(device_data - bench_data)
    return abs_err


def get_rel_err_origin(abs_err, b_value):
    rel_err_origin = np.abs(abs_err / b_value)
    return rel_err_origin


def get_max_abs_err(abs_err):
    max_abs_err = abs_err.max()
    bool_result = max_abs_err < 0.001
    return max_abs_err, bool_result


#相对误差最大值
def get_max_rel_err(rel_err):
    return np.max(rel_err) if np.max(rel_err) >= 0 else 0


#相对误差均值
def get_mean_rel_err(rel_err):
    non_negative_rel_err = rel_err[rel_err >= 0]
    return np.mean(non_negative_rel_err) if non_negative_rel_err.size > 0 else 0


def get_rel_err_ratio(rel_err, thresholding):
    if np.size(rel_err) == 0:
        ratio = 1
    else:
        ratio = np.divide(np.sum(rel_err < thresholding), np.size(rel_err))
    bool_result = ratio > (1 - thresholding)
    return ratio, bool_result


def get_finite_and_infinite_mask(bench_output, device_output):
    device_finite_mask = np.isfinite(device_output)
    bench_finite_mask = np.isfinite(bench_output.astype(device_output.dtype))
    both_finite_mask = np.logical_and(device_finite_mask, bench_finite_mask)
    inf_nan_mask = np.logical_not(both_finite_mask)
    return both_finite_mask, inf_nan_mask


def get_small_value_mask(abs_bench, both_finite_mask, small_value_threshold):
    small_value_mask = np.less_equal(abs_bench, small_value_threshold)
    small_value_mask = np.logical_and(small_value_mask, both_finite_mask)
    return small_value_mask


def get_abs_bench_with_eps(bench, dtype):
    abs_bench = np.abs(bench)
    eps = np.finfo(bench.dtype).eps if dtype != torch.bfloat16 else CompareConst.BFLOAT16_EPS
    abs_bench_with_eps = abs_bench + eps
    return abs_bench, abs_bench_with_eps


def check_inf_nan_value(inf_nan_mask, bench_output, device_output, dtype, rtol):
    '''
    新精度标准的绝对阈值法中，检查npu和golden输出的inf、nan是否一致
    输入：
        inf_nan_mask：npu输出和golden输出的inf、nan的mask
        bench_output：golden输出
        device_output：npu输出
        dtype：npu输出的dtype
    输出： 
        inf_nan_err_ratio：npu输出和golden输出的inf、nan不一致的比例
    '''
    _, abs_gpu_with_eps = get_abs_bench_with_eps(bench_output, dtype)
    golden_same_dtype = bench_output.astype(device_output.dtype)
    a_min = np.finfo(device_output.dtype).min if dtype != torch.bfloat16 else CompareConst.BFLOAT16_MIN
    a_max = np.finfo(device_output.dtype).max if dtype != torch.bfloat16 else CompareConst.BFLOAT16_MAX
    golden_clip = np.clip(golden_same_dtype, a_min, a_max)
    npu_clip = np.clip(device_output, a_min, a_max)
    clipped_abs_ae = np.abs(npu_clip - golden_clip)
    clipped_re = clipped_abs_ae / abs_gpu_with_eps
    pass_mask = np.less_equal(clipped_re, rtol)
    both_nan_mask = np.logical_and(np.isnan(device_output), np.isnan(golden_clip))
    pass_mask = np.logical_or(pass_mask, both_nan_mask)
    not_pass_mask = np.logical_not(pass_mask)
    not_pass_mask = np.logical_and(not_pass_mask, inf_nan_mask)

    inf_nan_err_cnt = np.sum(not_pass_mask)
    return 0 if np.sum(inf_nan_mask) == 0 else inf_nan_err_cnt / np.sum(inf_nan_mask)


def check_small_value(abs_err, small_value_mask, small_value_atol):
    '''
    新精度标准的绝对阈值法中，检查npu和golden正常值输出的绝对误差是否满足阈值
    输入：
        abs_err：npu输出和golden输出的绝对误差
        normal_value_mask：npu输出和golden输出的正常值mask
        atol：绝对误差的阈值
    输出： 
        abs_err_ratio：npu输出和golden输出的绝对误差不满足阈值的比例
    '''
    greater_mask = np.greater(abs_err, small_value_atol)
    err_mask = np.logical_and(greater_mask, small_value_mask)
    err_cnt = np.sum(err_mask)
    return 0 if np.sum(small_value_mask) == 0 else err_cnt / np.sum(small_value_mask)


def check_norm_value(normal_value_mask, rel_err, rtol):
    '''
    新精度标准的相对阈值法中，检查npu和golden小值域输出的相对误差是否满足阈值
    输入：
        rel_err：npu输出和golden输出的相对误差
        normal_value_mask：npu输出和golden输出的正常值mask
        rtol：相对误差的阈值
    输出： 
        rel_err_ratio：npu输出和golden输出的相对误差不满足阈值的比例
    '''
    err_mask = np.greater(rel_err, rtol)
    err_mask = np.logical_and(err_mask, normal_value_mask)
    err_cnt = np.sum(err_mask)
    return 0 if np.sum(normal_value_mask) == 0 else err_cnt / np.sum(normal_value_mask)


def get_ulp_err(bench_output, device_output, dtype):
    if is_dtype_fp8(dtype):
        return calc_ulp_err_fp8(bench_output, device_output)
    parameters = ULP_PARAMETERS.get(dtype)
    min_eb = parameters.get('min_eb')[0]
    exponent_num = parameters.get('exponent_num')[0]
    abs_bench = np.abs(bench_output)
    eb = np.where(abs_bench == 0, 0, np.floor(np.log2(abs_bench)))
    eb = np.maximum(eb, min_eb)

    if dtype == torch.float32:
        ulp_err = calc_ulp_err(bench_output, device_output, eb, exponent_num, np.float64)
    else:
        ulp_err = calc_ulp_err(bench_output, device_output, eb, exponent_num, np.float32)
    ulp_err = np.abs(ulp_err)
    return ulp_err


def calc_ulp_err_fp8(bench_output, device_output):
    # compute ulp error of FP8
    x = np.float64(bench_output)
    hi_fp8 = np.float64(device_output)

    ex = np.log2(abs(x) + 2**(-1000))
    ex[ex < -22] = -22
    exponent = np.floor(ex)  # Exponent

    eabs = np.abs(exponent)
    wm = np.zeros_like(x)                       # Mantissa width Init
    wm[eabs <= 15] = 1
    wm[eabs <= 7] = 2
    wm[eabs <= 3] = 3
    ulp_err = (hi_fp8 - x) * 2 ** (-exponent + wm)         # for wm = 1~3

    s_ex = ex * np.where(x >= 0, 1, -1)         
    eh = np.log2(abs(hi_fp8) + 2**(-1000))
    
    s_eh = eh * np.where(hi_fp8 >= 0, 1, -1)
    ulp_err1 = s_eh - s_ex                      # for wm = 0

    ulp_err[wm == 0] = ulp_err1[wm == 0]        # Merge 2 cases
    return ulp_err


def calc_ulp_err(bench_output, device_output, eb, exponent_num, data_type):
    return (device_output.astype(data_type) - bench_output).astype(data_type) * \
            np.exp2(-eb + exponent_num).astype(data_type)


def calc_ratio(x, y, dtype):
    """
    Calculate the ratio between NPU and GPU statistical values.

    Args:
        x (float): Statistical value from the NPU side
        y (float): Statistical value from the GPU side
        dtype: Data type used to determine the minimum error value

    Returns:
        float: The ratio of NPU to GPU statistical values

    Notes:
        - Takes absolute values of both x and y for calculation
        - Uses StandardConfig.get_minmum_err(dtype) to get minimum error for the specified dtype
        - Prevents division by zero by ensuring denominator is not less than minimum error
        - Returns |x| / max(|y|, minimum_error)
    """
    x, y = abs(x), abs(y)
    minmum_err = StandardConfig.get_minmum_err(dtype)
    err_y = max(y, minmum_err)
    return x / err_y


def compare_bool_tensor(bench_output, device_output):
    error_nums = (bench_output != device_output).sum()
    error_rate = float(error_nums / bench_output.size)
    result = CompareConst.PASS if error_rate == 0 else CompareConst.ERROR
    return error_rate, result, ""



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import multiprocessing
from dataclasses import dataclass
from functools import partial

import pandas as pd
from tqdm import tqdm

from msprobe.core.common.log import logger
from msprobe.core.common.utils import CompareException
from msprobe.core.common.const import CompareConst
from msprobe.core.common.exceptions import FileCheckException
from msprobe.core.compare.npy_compare import compare_ops_apply, get_error_flag_and_msg
from msprobe.core.compare.config import ModeConfig


@dataclass
class ComparisonResult:
    cos_result: list
    euc_dist_result: list
    max_err_result: list
    max_relative_err_result: list
    one_thousand_err_ratio_result: list
    five_thousand_err_ratio_result: list
    err_msgs: list


def _ms_graph_handle_multi_process(func, result_df, mode):
    process_num = max(int((multiprocessing.cpu_count() + 1) // 4), 1)
    df_chunk_size = len(result_df) // process_num
    if df_chunk_size > 0:
        df_chunks = [result_df.iloc[i:i + df_chunk_size] for i in range(0, len(result_df), df_chunk_size)]
    else:
        df_chunks = [result_df]

    results = []
    pool = multiprocessing.Pool(process_num)

    def err_call(args):
        logger.error('multiprocess compare failed! Reason: {}'.format(args))

    for df_chunk in df_chunks:
        result = pool.apply_async(func, args=(df_chunk, mode), error_callback=err_call)
        results.append(result)

    pool.close()

    try:
        final_results = [r.get(timeout=3600) for r in results]
    except Exception as e:
        logger.error(f"Task failed with exception: {e}")
        pool.terminate()
        raise CompareException(CompareException.MULTIPROCESS_ERROR) from e

    pool.join()
    return pd.concat(final_results, ignore_index=True)


def check_accuracy(cos, max_abs_err):
    if cos == CompareConst.SHAPE_UNMATCH:
        return CompareConst.ACCURACY_CHECK_UNMATCH
    if cos == CompareConst.NONE or max_abs_err == CompareConst.NONE:
        return CompareConst.NONE
    if cos == "N/A" or max_abs_err == "N/A":
        return CompareConst.ACCURACY_CHECK_NO
    try:
        cos, max_abs_err = float(cos), float(max_abs_err)
    except ValueError:
        logger.warning("Cosine or MaxAbsErr can not get float value.")
        return CompareConst.NONE
    if cos < CompareConst.COS_THRESHOLD and max_abs_err > CompareConst.MAX_ABS_ERR_THRESHOLD:
        return CompareConst.ACCURACY_CHECK_NO
    if cos < CompareConst.COS_MAX_THRESHOLD or max_abs_err > CompareConst.MAX_ABS_ERR_MAX_THRESHOLD:
        return CompareConst.ACCURACY_CHECK_NO
    return CompareConst.ACCURACY_CHECK_YES


class CompareRealData:
    def __init__(self, file_reader, mode_config: ModeConfig, cross_frame):
        self.file_reader = file_reader
        self.mode_config = mode_config
        self.cross_frame = cross_frame

    @staticmethod
    def read_dump_data(result_df):
        try:
            npu_dump_name_list = result_df.loc[0:, CompareConst.NPU_NAME].tolist()
            dump_tensor_pair_list = result_df.loc[0:, CompareConst.DATA_NAME].tolist()
            op_name_mapping_dict = {}
            for index, npu_dump_name in enumerate(npu_dump_name_list):
                dump_tensor_pair = dump_tensor_pair_list[index]
                op_name_mapping_dict[npu_dump_name] = dump_tensor_pair
            return op_name_mapping_dict
        except ValueError as e:
            logger.error('result dataframe is not found.')
            raise CompareException(CompareException.INVALID_DATA_ERROR) from e
        except KeyError as e:
            logger.error('result dataframe elements can not be access.')
            raise CompareException(CompareException.INVALID_KEY_ERROR) from e

    @staticmethod
    def _save_cmp_result(offset, result: ComparisonResult, result_df, lock):
        """
            Save comparison results into the result DataFrame with thread safety.
        Args:
            offset: offset for index
            result: data struct of ComparisonResult
            result_df: result of DataFrame
            lock: thread lock

        Returns:
            comparison results in DataFrame
        """

        lock.acquire()
        try:
            for i, cos_item in enumerate(result.cos_result):
                process_index = i + offset
                result_df.loc[process_index, CompareConst.COSINE] = cos_item
                result_df.loc[process_index, CompareConst.EUC_DIST] = result.euc_dist_result[i]
                result_df.loc[process_index, CompareConst.MAX_ABS_ERR] = result.max_err_result[i]
                result_df.loc[process_index, CompareConst.MAX_RELATIVE_ERR] = result.max_relative_err_result[i]
                result_df.loc[process_index, CompareConst.ONE_THOUSANDTH_ERR_RATIO] = (
                    result.one_thousand_err_ratio_result)[i]
                result_df.loc[process_index, CompareConst.FIVE_THOUSANDTHS_ERR_RATIO] = (
                    result.five_thousand_err_ratio_result)[i]
                result_df.loc[process_index, CompareConst.ERROR_MESSAGE] += result.err_msgs[i]
            return result_df
        except ValueError as e:
            logger.error('result dataframe is not found.')
            raise CompareException(CompareException.INVALID_DATA_ERROR) from e
        except IndexError as e:
            logger.error('result dataframe elements can not be access.')
            raise CompareException(CompareException.INDEX_OUT_OF_BOUNDS_ERROR) from e
        finally:
            lock.release()

    def compare_by_op(self, npu_op_name, bench_op_name, op_name_mapping_dict, input_param):
        """
        :param npu_op_name: excel中的NPU_Name，例如：MintFunctional.conv2d.0.forward.input.3.0
        :param bench_op_name: excel中的Bench_Name，例如：Functional.conv2d.0.forward.input.3.0
        :param op_name_mapping_dict: op_name和npy或pt文件的映射关系
        :param input_param: npu_path/bench_path/stack_path等参数
        :return: result_list，包含余弦相似度、最大绝对误差、最大相对误差、千分之一误差率、千分之五误差率和错误信息
        用于读取excel中的NPU_Name和Bench_Name，根据映射关系找到npy或pt文件，然后读取文件中的数据进行比较，计算余弦相似度、欧式距离
        最大绝对误差、最大相对误差、千分之一误差率、千分之五误差率并生成错误信息
        """
        relative_err, error_flag, err_msg = None, False, None

        data_name_pair = op_name_mapping_dict.get(npu_op_name)
        npu_data_name = data_name_pair[0]
        bench_data_name = data_name_pair[1]

        error_file = data_name_pair

        if str(npu_data_name) == CompareConst.NO_REAL_DATA_FLAG:  # 没有npu真实数据
            n_value, b_value, error_flag = CompareConst.NO_REAL_DATA, CompareConst.NO_REAL_DATA, True
            err_msg = "NPU does not have data file."
        elif str(bench_data_name) == CompareConst.NO_REAL_DATA_FLAG:  # 没有bench真实数据
            n_value, b_value, error_flag = CompareConst.NO_REAL_DATA, CompareConst.NO_REAL_DATA, True
            err_msg = "Bench does not have data file."
        elif str(bench_data_name) == CompareConst.N_A:  # bench没匹配
            n_value, b_value, error_flag = CompareConst.API_UNMATCH, CompareConst.API_UNMATCH, True
            err_msg = "Bench api/module unmatched."
        else:
            npu_dir = input_param.get(CompareConst.NPU_DUMP_DATA_DIR)
            bench_dir = input_param.get(CompareConst.BENCH_DUMP_DATA_DIR)
            try:
                n_value, b_value = self.file_reader(npu_dir, npu_data_name, bench_dir, bench_data_name,
                                                    self.cross_frame)
            except IOError as error:
                error_file = error.filename
                n_value, b_value = CompareConst.READ_NONE, CompareConst.READ_NONE
                error_flag = True
            except (FileCheckException, CompareException):
                error_file = data_name_pair
                n_value, b_value = CompareConst.READ_NONE, CompareConst.READ_NONE
                error_flag = True

        # 通过n_value, b_value同时得到错误标志和错误信息
        if not err_msg:
            n_value, b_value, error_flag, err_msg = get_error_flag_and_msg(n_value, b_value, error_flag=error_flag,
                                                                           error_file=error_file)

        result_list, err_msg = compare_ops_apply(n_value, b_value, error_flag, err_msg)

        if self.mode_config.fuzzy_match and npu_op_name != bench_op_name and bench_op_name != CompareConst.N_A:
            err_msg += " Fuzzy matching data, the comparison accuracy may be affected."
        result_list.append(err_msg)
        return result_list

    def compare_ops(self, idx, dump_path_dict, result_df, lock, input_param):
        cos_result = []
        euc_dist_result = []
        max_err_result = []
        max_relative_err_result = []
        one_thousand_err_ratio_result = []
        five_thousand_err_ratio_result = []
        err_mess = []

        is_print_compare_log = input_param.get("is_print_compare_log")

        for i in range(len(result_df)):
            npu_op_name = result_df.iloc[i, 0]
            bench_op_name = result_df.iloc[i, 1]
            if is_print_compare_log:
                logger.info("start compare: {}".format(npu_op_name))

            cos_sim, euc_dist, max_abs_err, max_relative_err, one_thousand_err_ratio, five_thousand_err_ratio, err_msg \
                = self.compare_by_op(npu_op_name, bench_op_name, dump_path_dict, input_param)

            if is_print_compare_log:
                if "does not have data file" in err_msg:
                    logger.info(f"[{npu_op_name}] Compare result: {err_msg} ")
                elif "Bench api/module unmatched" in err_msg:
                    logger.info(f"[{npu_op_name}] Compare result: {err_msg} ")
                else:
                    logger.info(
                        f"[{npu_op_name}] Compare result: cosine {cos_sim}, euc_dist {euc_dist}, "
                        f"max_abs_err {max_abs_err}, max_relative_err {max_relative_err}, "
                        f"one_thousand_err_ratio {one_thousand_err_ratio}, "
                        f"five_thousand_err_ratio {five_thousand_err_ratio}, {err_msg}"
                    )
            cos_result.append(cos_sim)
            euc_dist_result.append(euc_dist)
            max_err_result.append(max_abs_err)
            max_relative_err_result.append(max_relative_err)
            one_thousand_err_ratio_result.append(one_thousand_err_ratio)
            five_thousand_err_ratio_result.append(five_thousand_err_ratio)
            err_mess.append(err_msg)

        cr = ComparisonResult(
            cos_result=cos_result,
            euc_dist_result=euc_dist_result,
            max_err_result=max_err_result,
            max_relative_err_result=max_relative_err_result,
            one_thousand_err_ratio_result=one_thousand_err_ratio_result,
            five_thousand_err_ratio_result=five_thousand_err_ratio_result,
            err_msgs=err_mess
        )

        return self._save_cmp_result(idx, cr, result_df, lock)

    def do_multi_process(self, input_param, result_df):
        try:
            result_df = self._handle_multi_process(self.compare_ops, input_param, result_df,
                                                   multiprocessing.Manager().RLock())
            return result_df
        except ValueError as e:
            logger.error('result dataframe is not found.')
            raise CompareException(CompareException.INVALID_DATA_ERROR) from e

    def _handle_multi_process(self, func, input_param, result_df, lock):
        process_num = max(int((multiprocessing.cpu_count() + 1) // 4), 1)
        op_name_mapping_dict = self.read_dump_data(result_df)

        df_chunk_size = len(result_df) // process_num
        if df_chunk_size > 0:
            df_chunks = [result_df.iloc[i:i + df_chunk_size] for i in range(0, len(result_df), df_chunk_size)]
        else:
            df_chunks = [result_df]

        results = []
        pool = multiprocessing.Pool(process_num)

        def err_call(args):
            logger.error('multiprocess compare failed! Reason: {}'.format(args))

        progress_bar = tqdm(total=len(result_df), desc="API/Module Item Compare Process", unit="row", ncols=100)

        def update_progress(size, progress_lock, extra_param=None):
            with progress_lock:
                progress_bar.update(size)

        for process_idx, df_chunk in enumerate(df_chunks):
            idx = df_chunk_size * process_idx
            chunk_size = len(df_chunk)
            result = pool.apply_async(func,
                                      args=(idx, op_name_mapping_dict, df_chunk, lock, input_param),
                                      error_callback=err_call,
                                      callback=partial(update_progress, chunk_size, lock)
                                      )
            results.append(result)

        pool.close()

        try:
            final_results = [r.get(timeout=3600) for r in results]
        except Exception as e:
            logger.error(f"Task failed with exception: {e}")
            pool.terminate()
            raise CompareException(CompareException.MULTIPROCESS_ERROR) from e

        pool.join()
        return pd.concat(final_results, ignore_index=True)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.const import Const
from msprobe.core.dump.data_dump.data_processor.base import BaseDataProcessor
from msprobe.core.common.file_utils import check_file_or_directory_path
from msprobe.core.common.log import logger
from msprobe.core.common.exceptions import MsprobeException


class DataProcessorFactory:
    _data_processor = {}
    _module_processor = {}

    @classmethod
    def register_processor(cls, framework, task, processor_class):
        key = (framework, task)
        cls._data_processor[key] = processor_class

    @classmethod
    def register_module_processor(cls, framework, processor_class):
        cls._module_processor[framework] = processor_class

    @classmethod
    def get_module_processor(cls, framework):
        processor_class = cls._module_processor.get(framework)
        if not processor_class:
            raise ValueError(f"ModuleProcessor not found for framework: {framework}")
        return processor_class

    @classmethod
    def create_processor(cls, config, data_writer):
        cls.register_processors(config.framework)
        task = Const.KERNEL_DUMP if config.level == "L2" else config.task
        key = (config.framework, task)
        bench_path = getattr(config, "bench_path", None)
        if config.task == Const.TENSOR and bench_path is not None and not isinstance(bench_path, str):
            logger.error_log_with_exp("bench_path is invalid, it should be a string",
                                      MsprobeException(MsprobeException.INVALID_PARAM_ERROR))
        elif config.task == Const.TENSOR and bench_path is not None and isinstance(bench_path, str):
            check_file_or_directory_path(bench_path, True)
            processor_class = cls._data_processor.get(("pytorch", Const.DIFF_CHECK))
        else:
            processor_class = cls._data_processor.get(key)
        if not processor_class:
            raise ValueError(f"Processor not found for framework: {config.framework}, task: {config.task}")
        return processor_class(config, data_writer)

    @classmethod
    def register_processors(cls, framework):
        if framework == Const.PT_FRAMEWORK:
            from msprobe.core.dump.data_dump.data_processor.pytorch_processor import (
                StatisticsDataProcessor as PytorchStatisticsDataProcessor,
                TensorDataProcessor as PytorchTensorDataProcessor,
                DiffCheckDataProcessor as PytorchDiffCheckDataProcessor,
                KernelDumpDataProcessor as PytorchKernelDumpDataProcessor
            )
            from msprobe.pytorch.dump.module_dump.module_processor import ModuleProcessor
            cls.register_processor(Const.PT_FRAMEWORK, Const.STATISTICS, PytorchStatisticsDataProcessor)
            cls.register_processor(Const.PT_FRAMEWORK, Const.TENSOR, PytorchTensorDataProcessor)
            cls.register_processor(Const.PT_FRAMEWORK, Const.KERNEL_DUMP, PytorchKernelDumpDataProcessor)
            cls.register_processor(Const.PT_FRAMEWORK, Const.DIFF_CHECK, PytorchDiffCheckDataProcessor)
            cls.register_processor(Const.PT_FRAMEWORK, Const.STRUCTURE, BaseDataProcessor)
            cls.register_module_processor(Const.PT_FRAMEWORK, ModuleProcessor)
        elif framework == Const.MS_FRAMEWORK:
            from msprobe.core.dump.data_dump.data_processor.mindspore_processor import (
                StatisticsDataProcessor as MindsporeStatisticsDataProcessor,
                TensorDataProcessor as MindsporeTensorDataProcessor,
                KernelDumpDataProcessor as MindsporeKernelDumpDataProcessor
            )
            from msprobe.mindspore.dump.cell_processor import CellProcessor
            cls.register_processor(Const.MS_FRAMEWORK, Const.STATISTICS, MindsporeStatisticsDataProcessor)
            cls.register_processor(Const.MS_FRAMEWORK, Const.TENSOR, MindsporeTensorDataProcessor)
            cls.register_processor(Const.MS_FRAMEWORK, Const.KERNEL_DUMP, MindsporeKernelDumpDataProcessor)
            cls.register_processor(Const.MS_FRAMEWORK, Const.STRUCTURE, BaseDataProcessor)
            cls.register_module_processor(Const.MS_FRAMEWORK, CellProcessor)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------



import multiprocessing
import os

from msprobe.mindspore.api_accuracy_checker.data_manager import (DataManager, ResultCsvEntry, write_csv_header,
                                                                 get_result_csv_header, get_detail_csv_header,
                                                                 check_csv_header)
from msprobe.mindspore.common.log import logger


class MultiDataManager(DataManager):
    def __init__(self, csv_dir, result_csv_path, shared_is_first_write):
        super().__init__(csv_dir, result_csv_path)

        # 使用共享的 is_first_write 变量来控制表头写入
        self.shared_is_first_write = shared_is_first_write
        # 创建锁对象，确保线程安全
        self.lock = multiprocessing.Lock()

    def save_results(self, api_name_str):
        """保存结果，线程安全操作"""

        with self.lock:  # 确保保存操作不会被多个进程同时进行
            if self.is_first_write and self.shared_is_first_write.value:
                self.shared_is_first_write.value = False
                self.is_first_write = False  # 写入后标记为 False，避免重复写入表头
                # 直接写入表头
                logger.info("Writing CSV headers for the first time.")
                write_csv_header(self.detail_out_path, get_detail_csv_header)
                write_csv_header(self.result_out_path, get_result_csv_header)

            """写入详细输出和结果摘要并清理结果"""
            self.to_detail_csv(self.detail_out_path)
            logger.debug(f"Detailed output for {api_name_str} written to {self.detail_out_path}.")

            self.to_result_csv(self.result_out_path)
            logger.debug(f"Result summary for {api_name_str} written to {self.result_out_path}.")

            # 清理记录，准备下一次调用
            self.clear_results()

    def clear_results(self):
        """清空 self.results 数据，线程安全操作"""
        logger.debug("Clearing results data.")
        self.results.clear()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import csv

from msprobe.core.common.const import Const, CompareConst
from msprobe.core.common.file_utils import FileOpen, create_directory, write_csv, read_csv
from msprobe.core.common.utils import add_time_as_suffix, MsprobeBaseException
from msprobe.mindspore.api_accuracy_checker.base_compare_algorithm import compare_algorithms
from msprobe.core.common.file_utils import check_file_or_directory_path
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.common.const import MsCompareConst


class ResultCsvEntry:
    def __init__(self) -> None:
        self.forward_pass_status = None
        self.backward_pass_status = None
        self.forward_err_msg = ""
        self.backward_err_msg = ""
        self.overall_err_msg = None


def write_csv_header(csv_path, header_func):
    """如果是第一次写入，则写入 CSV 表头"""
    header = header_func()  # 获取表头
    logger.debug(f"Writing CSV header: {header}")
    write_csv([header], csv_path, mode="a+")


def get_result_csv_header():
    """获取结果 CSV 文件的表头"""
    return [
        MsCompareConst.DETAIL_CSV_API_NAME,
        MsCompareConst.RESULT_CSV_FORWARD_TEST_SUCCESS,
        MsCompareConst.RESULT_CSV_BACKWARD_TEST_SUCCESS,
        MsCompareConst.DETAIL_CSV_MESSAGE,
    ]


def get_detail_csv_header():
    """获取详细 CSV 文件的表头"""
    detail_csv_header_basic_info = [
        MsCompareConst.DETAIL_CSV_API_NAME,
        MsCompareConst.DETAIL_CSV_BENCH_DTYPE,
        MsCompareConst.DETAIL_CSV_TESTED_DTYPE,
        MsCompareConst.DETAIL_CSV_SHAPE,
    ]
    detail_csv_header_compare_result = list(compare_algorithms.keys())
    detail_csv_header_status = [
        MsCompareConst.DETAIL_CSV_PASS_STATUS,
        MsCompareConst.DETAIL_CSV_MESSAGE,
    ]
    return detail_csv_header_basic_info + detail_csv_header_compare_result + detail_csv_header_status


def check_csv_header(headers, required_constants, csv_path):
    """校验 CSV 文件表头是否包含所有必需的常量"""
    missing_constants = [const for const in required_constants if not any(const in header for header in headers)]

    if missing_constants:
        raise MsprobeBaseException(
            MsprobeBaseException.MISSING_HEADER_ERROR,
            f"{csv_path} 缺少以下必需的表头字段: {missing_constants}"
        )
    return True


class DataManager:
    def __init__(self, csv_dir, result_csv_path):
        self.results = {}
        self.results_exception_skip = {}
        self.is_first_write = True  # 标记用于添加表头
        self.csv_dir = csv_dir
        self.api_names_set = set()  # 存储已经出现的 API 名称的集合
        # 如果传入了 result_csv_path，则启用断点续检
        if result_csv_path:
            self.resume_from_last_csv(result_csv_path)
            self.initialize_api_names_set(result_csv_path)
        else:
            # 默认情况下，设置输出路径为空，等待首次写入时初始化
            self.result_out_path = os.path.join(self.csv_dir, add_time_as_suffix(MsCompareConst.RESULT_CSV_FILE_NAME))
            self.detail_out_path = os.path.join(
                self.csv_dir,
                os.path.basename(self.result_out_path).replace("result", "details")
            )

            if self.detail_out_path and os.path.exists(self.detail_out_path):
                check_file_or_directory_path(self.detail_out_path)

            if self.result_out_path and os.path.exists(self.result_out_path):
                check_file_or_directory_path(self.result_out_path)

    def initialize_api_names_set(self, result_csv_path):
        """读取现有的 CSV 文件并存储已经出现的 API 名称到集合中"""
        # 使用新的 read_csv 函数读取数据
        csv_data = read_csv(result_csv_path, as_pd=False)

        # 读取标题行
        headers = csv_data[0] if csv_data else []  # 如果文件为空，则 headers 会为空

        # 使用提取的表头校验函数
        if check_csv_header(headers, get_result_csv_header(), result_csv_path):

            # 获取 "API Name" 列的索引
            api_name_index = None
            for i, header in enumerate(headers):
                if MsCompareConst.DETAIL_CSV_API_NAME in header:  # CSV 文件的标题行包含了字节顺序标记,所以使用通过包含方式来查找
                    api_name_index = i
                    break

            if api_name_index is None:
                logger.warning(f"{result_csv_path} No column contains 'API Name'.")
                return

            # 读取每一行的 API 名称
            for row in csv_data[1:]:  # 跳过标题行，从第二行开始
                if row and len(row) > api_name_index:
                    api_name = row[api_name_index]
                    if api_name:
                        self.api_names_set.add(api_name)

            logger.debug(f"Initialized API names set from existing CSV: {self.api_names_set}")

    def is_unique_api(self, api_name):
        """检查 API 名称是否唯一，如果已经存在则返回 False，否则加入集合并返回 True"""
        if api_name in self.api_names_set:
            return False
        self.api_names_set.add(api_name)
        return True

    def resume_from_last_csv(self, result_csv_path):
        """从上次运行的 result_csv_path 恢复断点"""
        # 获取上次的目录路径
        last_dir = os.path.dirname(result_csv_path)

        # 设置当前目录和输出路径，确保在首次写入时使用
        self.csv_dir = last_dir
        self.detail_out_path = os.path.join(last_dir, os.path.basename(result_csv_path).replace("result", "details"))
        if self.detail_out_path and os.path.exists(self.detail_out_path):
            check_file_or_directory_path(self.detail_out_path)
        self.result_out_path = result_csv_path
        self.is_first_write = False

    def save_results(self, api_name_str):
        if self.is_first_write:
            # 直接写入表头
            logger.info("Writing CSV headers for the first time.")
            write_csv_header(self.detail_out_path, get_detail_csv_header)
            write_csv_header(self.result_out_path, get_result_csv_header)
            self.is_first_write = False  # 写入后标记为 False，避免重复写入表头

        """写入详细输出和结果摘要并清理结果"""
        logger.debug("Starting to write detailed output to CSV.")
        self.to_detail_csv(self.detail_out_path)
        logger.debug(f"Detailed output for {api_name_str} written to {self.detail_out_path}.")

        logger.debug("Starting to write result summary to CSV.")
        self.to_result_csv(self.result_out_path)
        logger.debug(f"Result summary for {api_name_str} written to {self.result_out_path}.")

        # 清理记录，准备下一次调用
        self.clear_results()

    def record(self, output_list):
        if output_list is None:
            return
        for output in output_list:
            api_real_name, forward_or_backward, basic_info, compare_result_dict = output
            key = (api_real_name, forward_or_backward)
            if key not in self.results:
                self.results[key] = []
            self.results[key].append((basic_info, compare_result_dict))
            logger.debug(f"Updated self.results for key {key}: {self.results[key]}")
        logger.debug(f"Complete self.results after recording: {self.results}")

    def record_exception_skip(self, api_name, forward_or_backward, err_msg):
        '''
            record exception_skip information into self.record_exception_skip.
            self.record_exception_skip: dict{str: dict{"forward": str/None, "backward": str/None}}
            string in key is api_name, string in value is err_msg
        '''
        if api_name not in self.results_exception_skip:
            self.results_exception_skip[api_name] = {Const.FORWARD: None, Const.BACKWARD: None}
        self.results_exception_skip[api_name][forward_or_backward] = err_msg

    def clear_results(self):
        """清空 self.results 数据"""
        logger.debug("Clearing self.results data.")
        self.results.clear()
        self.results_exception_skip.clear()

    def to_detail_csv(self, csv_path):
        logger.debug("Preparing detail CSV headers and rows.")
        detail_csv = []

        detail_csv_header_compare_result = list(compare_algorithms.keys())

        for _, results in self.results.items():
            for res in results:
                basic_info, compare_result_dict = res
                csv_row_basic_info = [
                    basic_info.api_name,
                    basic_info.bench_dtype,
                    basic_info.tested_dtype,
                    basic_info.shape
                ]
                csv_row_compare_result = [
                    compare_result_dict.get(algorithm_name).compare_value
                    for algorithm_name in detail_csv_header_compare_result
                ]
                csv_row_status = [basic_info.status, basic_info.err_msg]
                csv_row = csv_row_basic_info + csv_row_compare_result + csv_row_status
                detail_csv.append(csv_row)
                logger.debug(f"Detail CSV row added: {csv_row}")

        logger.debug(f"Writing detail CSV to {csv_path}.")
        write_csv(detail_csv, csv_path, mode="a+")
        logger.debug(f"Detail CSV written successfully to {csv_path}.")

    def to_result_csv(self, csv_path):
        '''
            depend on both self.results and self.results_exception_skip
        '''
        logger.debug("Preparing result CSV data.")
        result_csv = []

        result_csv_dict = {}
        for key, results in self.results.items():
            api_real_name, forward_or_backward = key
            pass_status = CompareConst.PASS
            overall_err_msg = ""

            for res in results:
                basic_info, _ = res
                if basic_info.status != CompareConst.PASS:
                    pass_status = CompareConst.ERROR
                overall_err_msg += basic_info.err_msg

            overall_err_msg = "" if pass_status == CompareConst.PASS else overall_err_msg

            if api_real_name not in result_csv_dict:
                result_csv_dict[api_real_name] = ResultCsvEntry()
            if forward_or_backward == Const.FORWARD:
                result_csv_dict[api_real_name].forward_pass_status = pass_status
                result_csv_dict[api_real_name].forward_err_msg = overall_err_msg
            else:
                result_csv_dict[api_real_name].backward_pass_status = pass_status
                result_csv_dict[api_real_name].backward_err_msg = overall_err_msg

        for api_name, entry in result_csv_dict.items():
            overall_err_msg = "" if (entry.forward_pass_status == CompareConst.PASS and
                                     entry.backward_pass_status == CompareConst.PASS) else \
                entry.forward_err_msg + entry.backward_err_msg
            row = [
                api_name,
                entry.forward_pass_status,
                entry.backward_pass_status,
                overall_err_msg
            ]
            # change row if this api has exception_skip information
            if api_name in self.results_exception_skip:
                if self.results_exception_skip[api_name][Const.FORWARD] is not None:
                    row[1] = CompareConst.SKIP
                    row[-1] += self.results_exception_skip[api_name][Const.FORWARD]
                if self.results_exception_skip[api_name][Const.BACKWARD] is not None:
                    row[2] = CompareConst.SKIP
                    row[-1] += self.results_exception_skip[api_name][Const.BACKWARD]
                del self.results_exception_skip[api_name]
            result_csv.append(row)
            logger.debug(f"Result CSV row added: {row}")
        for api_name in self.results_exception_skip:
            current_exception_skip = self.results_exception_skip[api_name]
            forward_status = None
            backward_status = None
            err_msg = ""
            if current_exception_skip[Const.FORWARD] is not None:
                forward_status = CompareConst.SKIP
                err_msg += current_exception_skip[Const.FORWARD]
            if current_exception_skip[Const.BACKWARD] is not None:
                backward_status = CompareConst.SKIP
                err_msg += current_exception_skip[Const.BACKWARD]
            row = [api_name, forward_status, backward_status, err_msg]
            result_csv.append(row)

        write_csv(result_csv, csv_path, mode="a+")
        logger.debug(f"Result CSV written successfully to {csv_path}.")

        # 设置标记为 False，防止后续重复添加表头
        self.is_first_write = False



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

import torch

from msprobe.core.common.utils import logger, CompareException
from msprobe.core.common.file_utils import FileChecker, FileCheckConst
from msprobe.pytorch.common.utils import load_pt


def read_pt_data(dir_path, file_name):
    if not file_name:
        return None

    data_path = os.path.join(dir_path, file_name)
    path_checker = FileChecker(data_path, FileCheckConst.FILE, FileCheckConst.READ_ABLE, FileCheckConst.PT_SUFFIX)
    data_path = path_checker.common_check()
    try:
        # detach because numpy can not process gradient information
        data_value = load_pt(data_path, to_cpu=True).detach()
    except RuntimeError as e:
        # 这里捕获 load_pt 中抛出的异常
        data_path_file_name = os.path.basename(data_path)
        logger.error(f"Failed to load the .pt file at {data_path_file_name}.")
        raise CompareException(CompareException.INVALID_FILE_ERROR) from e
    except AttributeError as e:
        # 这里捕获 detach 方法抛出的异常
        logger.error(f"Failed to detach the loaded tensor.")
        raise CompareException(CompareException.DETACH_ERROR) from e
    if data_value.dtype == torch.bfloat16:
        data_value = data_value.to(torch.float32)
    data_value = data_value.numpy()
    return data_value



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the mapping function.
"""
import os
import csv

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.utils import check_file_size


def _handle_csv_object(csv_object: any, mapping_file_path: str) -> dict:
    hash_to_file_name_map = {}
    for item in csv_object:
        if len(item) == 2:
            hash_to_file_name_map[item[0]] = item[1]
        else:
            log.print_error_log(
                'The content (%s) of the mapping file "%r" is invalid.' % (item, mapping_file_path))
    return hash_to_file_name_map


def read_mapping_file(mapping_file_path: str) -> dict:
    """
    Read mapping file
    :param mapping_file_path: mapping file path
    :return: hash_to_file_name_map
    """
    hash_to_file_name_map = {}
    if not os.path.isfile(mapping_file_path):
        return hash_to_file_name_map
    check_file_size(mapping_file_path, ConstManager.ONE_HUNDRED_MB)
    try:
        with open(mapping_file_path, "r") as mapping_file:
            csv_object = csv.reader(mapping_file)
            return _handle_csv_object(csv_object, mapping_file_path)
    except csv.Error:
        log.print_error_log('Failed to read csv object. The content of the mapping file "%r" is invalid.'
                            % mapping_file_path)
    except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
        log.print_open_file_error(mapping_file_path, error)
    finally:
        pass
    return hash_to_file_name_map



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2026 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from enum import Enum, auto


# 统计值类型枚举
class ValType(Enum):
    NUM = auto()
    NAN = auto()
    INF = auto()
    NEG_INF = auto()
    DEVICE = auto()
    NA = auto()
    OTHER = auto()


ALL_TYPES = [
    ValType.NUM,
    ValType.NAN,
    ValType.INF,
    ValType.NEG_INF,
    ValType.DEVICE,
    ValType.NA,
    ValType.OTHER,
]



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
Quant Filter class, process the fusion json file
"""
import re

from msprobe.msaccucmp.cmp_utils import log


class QuantFilter:
    """
    the class for recognize and filter the quant/dequant ops
    """

    NORMAL_OP = 0
    MIDDLE_OP = 1
    QUANT_OP = 2
    DEQUANT_OP = 3
    QUANT_DEQUANT_OP = 5

    def __init__(self: any, op_list: list) -> None:
        """
        :param op_list: the list of parsed ops
        """
        self._op_list = op_list
        self._op_output_type_map = {}  # structure like {op_name: [out1_type, out2_type]}
        self._quant_name_pattern = r"_quant_layer|/AscendQuant|/AscendWeightQuant|\.quant|\.weight_quant"
        self._dequant_name_pattern = \
            r"_dequant_layer|_anti_quant_layer|/AscendDequant|/AntiQuant|\.dequant|\.anti_quant"

    @staticmethod
    def _check_out_type(op_name: str, name_type: int, in_pairs: bool, output_node: any) -> int:
        data_type = output_node.data_type
        _out_type = QuantFilter._match_existing_type(name_type, data_type, in_pairs)

        # Not matched type, treat as a normal op
        if _out_type is None:
            log.print_warn_log("[{}] this op looks like a quantization op, but it does not have proper output type"
                               .format(op_name))
            if in_pairs:
                _out_type = QuantFilter.MIDDLE_OP
            else:
                _out_type = QuantFilter.NORMAL_OP
        return _out_type

    @staticmethod
    def _match_existing_type(name_type: int, data_type: str, in_pairs: bool) -> int:
        _matched_type = None
        if name_type == QuantFilter.NORMAL_OP:
            if in_pairs:
                _matched_type = QuantFilter.MIDDLE_OP
            else:
                _matched_type = QuantFilter.NORMAL_OP
        elif name_type == QuantFilter.QUANT_OP:
            if data_type == "DT_INT8":
                _matched_type = QuantFilter.QUANT_OP
        elif name_type == QuantFilter.DEQUANT_OP:
            if data_type == "DT_FLOAT16":
                _matched_type = QuantFilter.DEQUANT_OP
        elif name_type == QuantFilter.QUANT_DEQUANT_OP:
            # 5--quant & dequant
            if data_type == "DT_INT8":
                _matched_type = QuantFilter.QUANT_OP
            if data_type == "DT_FLOAT16":
                _matched_type = QuantFilter.DEQUANT_OP

        return _matched_type

    def process_filtering(self: any) -> None:
        """
        check through the op list, mark in attr.is_quant_filter the ops should be filtered.
        """
        self._op_output_type_map = {}

        # fetch ops in order
        for op in self._op_list:

            _op_name = op.op_name

            # check name type using regex
            _name_type = self._check_name_type(op)

            # check inputs to determine if in middle of pairs
            _in_pairs = self._check_in_pairs(op)

            # store the output type of each output desc
            self._op_output_type_map[_op_name] = []

            for output in op.output_desc:
                # combine the conditions above to determine filtering or not
                _out_type = QuantFilter._check_out_type(_op_name, _name_type, _in_pairs, output)
                self._op_output_type_map[_op_name].append(_out_type)

            # store filter result in result set
            self._add_filter_list(op, _in_pairs)

    def get_out_type_map(self: any) -> dict:
        """
        get output type map

        :return: output map
        """
        return self._op_output_type_map

    def _check_in_pairs(self: any, op: any) -> bool:
        _inputs = op.input_list

        _is_in_pair = False
        for _input in _inputs:
            if ':' not in _input:
                continue
            last_colon_index = _input.rfind(':')
            _input_name = _input[:last_colon_index]
            if _input[last_colon_index + 1:].isdigit():
                _input_out_index = int(_input[last_colon_index + 1:])
            else:
                log.print_warn_log("[{}] the input operator of op with index {} is invalid."
                                   .format(op.op_name, _input))
                continue
            _out_list = self._op_output_type_map.get(_input_name)
            if _out_list is None or len(_out_list) <= _input_out_index:
                log.print_warn_log("[{}] the input operator of op {} does not exist in previous graph."
                                   .format(op.op_name, _input_name))
                continue
            _input_out_type = _out_list[_input_out_index]
            if _input_out_type in (QuantFilter.QUANT_OP, QuantFilter.MIDDLE_OP):
                _is_in_pair = True
                break

        return _is_in_pair

    def _add_filter_list(self: any, op: any, in_pairs: bool) -> None:
        _outputs = self._op_output_type_map.get(op.op_name)
        _is_dequant = True
        for _output in _outputs:
            if _output != QuantFilter.DEQUANT_OP:
                _is_dequant = False
                break
        _need_filter = in_pairs and not _is_dequant

        op.attr.quant_filter = _need_filter

    def _check_name_type(self, op) -> int:
        """
        check the op type with name, using regex

        :param op: the op name need to be checked
        :return: type of this op, Quant=2/Dequant=3/Quant&Dequant=5/Normal=0
        """

        _op_type = QuantFilter.NORMAL_OP

        # All the keywords of quant/dequant ops:
        # caffe: "_quant_layer", "_dequant_layer", "_anti_quant_layer"
        # tensorflow: "/AscendQuant", "/AscendWeightQuant", "/AscendDequant", "/AntiQuant"
        # onnx: ".quant", ".weight_quant", ".dequant", ".anti_quant"

        # find quant keywords
        if re.search(self._quant_name_pattern, op.op_name):
            _op_type += QuantFilter.QUANT_OP
        elif re.search(self._quant_name_pattern, ''.join(op.attr.original_op_names)):
            _op_type += QuantFilter.QUANT_OP

        # find dequant keywords
        if re.search(self._dequant_name_pattern, op.op_name):
            _op_type += QuantFilter.DEQUANT_OP
        elif re.search(self._dequant_name_pattern, ''.join(op.attr.original_op_names)):
            _op_type += QuantFilter.DEQUANT_OP

        return _op_type



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import inspect
import os
import re

import torch
import torch.distributed as dist
import torch.nn as nn

from msprobe.core.common.const import MonitorConst
from msprobe.core.common.file_utils import load_yaml
from msprobe.pytorch.monitor.module_metric import get_metrics, get_summary_writer_tag_name
from msprobe.pytorch.common.log import logger

try:
    import torch_npu
except ImportError:
    pass

RANK = None

OpsPath = os.path.join(os.path.dirname(__file__), "distributed_ops.yaml")
WrapDistributedOps = load_yaml(OpsPath).get("distributed", [])

StackBlackListPath = os.path.join(os.path.dirname(__file__), "stack_blacklist.yaml")
StackBlackList = load_yaml(StackBlackListPath).get("stack", [])
MAX_STRING_LENGTH = 1000

distributed_func = {}
for f in dir(dist):
    distributed_func[f] = getattr(dist, f)

ORIGIN_WAIT = getattr(dist.Work, 'wait')
PENDING_ASYNC_CC_BY_HANDLE = {}


def get_distributed_ops():
    global WrapDistributedOps
    _all_distributed_ops = dir(dist)
    return set(WrapDistributedOps) & set(_all_distributed_ops)


class DistributedOPTemplate(nn.Module):
    def __init__(self, op_name, pre_hooks, post_hooks):
        super(DistributedOPTemplate, self).__init__()
        self.op_name_ = str(op_name)
        self.__name__ = self.op_name_
        self.cc_hooks = []
        for pre_hook in pre_hooks:
            handle = self.register_forward_pre_hook(pre_hook, with_kwargs=True)
            self.cc_hooks.append(handle)
        for hook in post_hooks:
            handle = self.register_forward_hook(hook, with_kwargs=True)
            self.cc_hooks.append(handle)

    def forward(self, *args, **kwargs):
        return distributed_func.get(self.op_name_)(*args, **kwargs)


class ApiRegistry:
    def __init__(self):
        self.distributed_attr_origin = {}
        self.distributed_attr_hooked = {}

    @staticmethod
    def store_ori_attr(ori_api_group, api_list, api_ori_attr):
        for api in api_list:
            if '.' in api:
                sub_module_name, sub_op = api.rsplit('.', 1)
                sub_module = getattr(ori_api_group, sub_module_name)
                api_ori_attr[api] = getattr(sub_module, sub_op)
            else:
                api_ori_attr[api] = getattr(ori_api_group, api)

    @staticmethod
    def set_api_attr(api_group, attr_dict):
        for cc_api_name, cc_api_entry_func in attr_dict.items():
            if '.' in cc_api_name:
                sub_module_name, sub_op = cc_api_name.rsplit('.', 1)
                sub_module = getattr(api_group, sub_module_name, None)
                if sub_module is not None:
                    setattr(sub_module, sub_op, cc_api_entry_func)
            else:
                setattr(api_group, cc_api_name, cc_api_entry_func)

    @staticmethod
    def redirect_wait():
        global ORIGIN_WAIT
        global PENDING_ASYNC_CC_BY_HANDLE

        def wrapped_wait(work):
            def wrapped_wait(*args, **kwargs):
                ORIGIN_WAIT(*args, **kwargs)
                if args[0] in PENDING_ASYNC_CC_BY_HANDLE:
                    store_func = PENDING_ASYNC_CC_BY_HANDLE.pop(args[0])
                    store_func()

            return wrapped_wait

        dist.Work.wait = wrapped_wait(dist.Work)

    def redirect_api(self):
        self.set_api_attr(dist, self.distributed_attr_hooked)
        self.set_api_attr(dist.distributed_c10d, self.distributed_attr_hooked)
        self.redirect_wait()

    def restore_api(self):
        self.set_api_attr(dist, self.distributed_attr_origin)
        self.set_api_attr(dist.distributed_c10d, self.distributed_attr_origin)
        setattr(dist.Work, 'wait', ORIGIN_WAIT)

    def initialize_hook(self, pre_hooks, post_hooks):
        self.store_ori_attr(dist, get_distributed_ops(), self.distributed_attr_origin)
        cc_hooks = []
        for op_name in get_distributed_ops():
            self.distributed_attr_hooked[op_name] = DistributedOPTemplate(op_name, pre_hooks, post_hooks)
            cc_hooks.extend(self.distributed_attr_hooked[op_name].cc_hooks)
        return cc_hooks


def get_process_group(process_group):
    return (
        process_group
        if isinstance(process_group, dist.ProcessGroup)
        else dist.GroupMember.WORLD
    )


def stack_filter(stack):
    if len(stack) > MAX_STRING_LENGTH:
        logger.warning(f'The character string contains more than {MAX_STRING_LENGTH}. re match is skipped.')
    for pattern in StackBlackList:
        if re.search(pattern, stack):
            return False
    return True


def get_callstack():
    callstack = []
    for (_, path, line, func, _, _) in inspect.stack():
        stack_line = f'{path}[{line}]'
        if stack_filter(stack_line):
            callstack.append(stack_line + '   ' + func)
    return callstack


@torch.no_grad()
def op_aggregate(op, tensorlist):
    if isinstance(tensorlist, torch.Tensor):
        return tensorlist
    if not tensorlist:
        return torch.tensor(torch.nan)
    if op == 'min':
        return min(tensorlist)
    if op == 'max':
        return max(tensorlist)
    if op == 'norm':
        return sum(tensorlist)
    if op == 'zeros':
        return sum(tensorlist) / len(tensorlist)
    if op == 'nans':
        return sum(tensorlist)
    if op == 'mean':
        return sum(tensorlist) / len(tensorlist)
    return torch.tensor(torch.nan)


def update_data(old, new):
    for tag, op2tensor in new.items():
        if tag not in old:
            old[tag] = {}
        for op, tensor in op2tensor.items():
            if op not in old[tag]:
                old[tag][op] = [tensor]
            else:
                old[tag][op].append(tensor)
    return old


def is_target_line(codeline):
    if codeline == []:
        return True
    stack = get_callstack()
    whole_stack = ';'.join(stack)
    if len(whole_stack) > MAX_STRING_LENGTH:
        logger.warning(f'The character string contains more than {MAX_STRING_LENGTH}. re match is skipped.')
    for pattern in codeline:
        if re.search(pattern, whole_stack):
            return True
    return False


@torch.no_grad()
def catch_data(cc_context, cc_name, ops, args, prefix):
    tensor_args = {}
    for arg in args:
        if isinstance(arg, torch.Tensor):
            key = get_summary_writer_tag_name(cc_name, f'{prefix}_{len(tensor_args)}', RANK)
            tensor_args[key] = arg
        elif isinstance(arg, list):
            if isinstance(arg[0], torch.Tensor):
                stacked_arg = torch.stack(arg)
            elif isinstance(arg[0], dist.P2POp):
                stacked_arg = torch.stack([op.tensor for op in arg])
            key = get_summary_writer_tag_name(cc_name, f'{prefix}_{len(tensor_args)}', RANK)
            tensor_args[key] = stacked_arg

    new_data = get_metrics(ops, tensor_args, 1e-8)
    cc_context.data = update_data(cc_context.data, new_data)


def create_async_callback_func(context, cc_name, ops, args, prefix):
    def store_data():
        catch_data(context, cc_name, ops, args, prefix)

    return store_data


def create_hooks(context, monitor):
    def cc_log_hook(module, args, kwargs):
        stack = ';'.join(get_callstack())
        monitor.cc_logged_stack[module.op_name_].add(stack)
        return

    def cc_pre_hook(module, args, kwargs):
        if not is_target_line(monitor.cc_codeline):
            return
        args = args + tuple(kwargs.values())
        catch_data(context[module.op_name_], module.op_name_, monitor.ops, args, MonitorConst.PREFIX_PRE)
        return

    def cc_hook(module, args, kwargs, out=None):
        if not is_target_line(monitor.cc_codeline):
            return out
        args = args + tuple(kwargs.values())
        if out:  # async
            if isinstance(out, dist.Work):
                PENDING_ASYNC_CC_BY_HANDLE[out] = create_async_callback_func(
                    context[module.op_name_],
                    module.op_name_,
                    monitor.ops, args,
                    MonitorConst.PREFIX_POST
                )
            elif isinstance(out, list):  # batch_isend_irecv
                for out_element in out:
                    PENDING_ASYNC_CC_BY_HANDLE[out_element] = create_async_callback_func(
                        context[module.op_name_],
                        module.op_name_,
                        monitor.ops, args,
                        MonitorConst.PREFIX_POST
                    )
            return out
        catch_data(context[module.op_name_], module.op_name_, monitor.ops, args, MonitorConst.PREFIX_POST)
        return out

    global RANK
    pre_hooks = []
    hooks = []
    RANK = dist.get_rank()
    if dist.is_initialized() and RANK not in monitor.module_rank_list and monitor.module_rank_list != []:
        return [pre_hooks, hooks]

    if monitor.cc_log_only:
        pre_hooks.append(cc_log_hook)
        return [pre_hooks, hooks]

    if monitor.cc_pre_hook:
        pre_hooks.append(cc_pre_hook)

    hooks.append(cc_hook)

    return [pre_hooks, hooks]


api_register = ApiRegistry()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import array
import os
import time

import pandas as pd

try:
    import torch
except ImportError:
    torch_available = False
else:
    torch_available = True

from msprobe.core.common.const import Const, FileCheckConst, CompareConst
from msprobe.core.common.exceptions import FileCheckException
from msprobe.core.common.file_utils import (
    check_and_get_real_path,
    check_if_valid_dir_pattern_path,
    read_csv,
    save_excel,
    create_directory,
    get_file_content_bytes
)
from msprobe.core.common.log import logger
from msprobe.core.compare.npy_compare import get_error_flag_and_msg, compare_ops_apply


MIN_ROW_COUNT_OF_CSV = 2
COL_COUNT_OF_CSV = 15
ROW_NUMBER_OF_FIRST_STAT = 1
DEVICE_AND_PID_COL_NUM = 0
EXECUTION_COUNT_COL_NUM = 1
OP_NAME_COL_NUM = 2
OP_ID_COL_NUM = 4
IO_TYPE_COL_NUM = 5
TENSOR_INDEX_COL_NUM = 6
DTYPE_COL_NUM = 7
SHAPE_COL_NUM = 9
MAX_VALUE_COL_NUM = 10
MIN_VALUE_COL_NUM = 11
MEAN_VALUE_COL_NUM = 12
NORM_VALUE_COL_NUM = 13
TENSOR_PATH_COL_NUM = 14

ATTR_VERSION = "$Version"
ATTR_END = "$End"
ATTR_OBJECT_LENGTH = "$Object.Length"
ATTR_OBJECT_COUNT = "$Object.Count"
ATTR_OBJECT_PREFIX = "$Object."

COMMON_HEADER = [
    CompareConst.ATB_DATA_NAME, CompareConst.DEVICE_AND_PID,
    CompareConst.EXECUTION_COUNT, CompareConst.DATA_TYPE, CompareConst.DATA_SHAPE
]

COMMON_HEADER_STATS = [Const.MAX, Const.MIN, Const.MEAN, Const.NORM]

STATS_SPECIAL_HEADER = [
    'Max Diff', 'Min Diff', 'Mean Diff', 'Norm Diff',
    'Relative Err of Max(%)', 'Relative Err of Min(%)',
    'Relative Err of Mean(%)', 'Relative Err of Norm(%)'
]

TENSOR_SPECIAL_HEADER = [
    'Cosine', 'Euc Distance', 'Max Absolute Err', 'Max Relative Err',
    'One Thousandth Err Ratio', 'Five Thousandth Err Ratio'
]


def compare_atb_mode(args):
    """
    Entry point used by the CLI to trigger ATB model accuracy compare.
    """
    args.golden_path = check_and_get_real_path(args.golden_path, FileCheckConst.READ_ABLE)
    args.target_path = check_and_get_real_path(args.target_path, FileCheckConst.READ_ABLE)
    if os.path.isfile(args.golden_path) != os.path.isfile(args.target_path):
        logger.error('The golden_path and target_path should be both files or directories')
        raise FileCheckException(FileCheckException.ILLEGAL_PATH_ERROR)
    check_if_valid_dir_pattern_path(args.output_path)
    args.output_path = check_and_get_real_path(args.output_path, FileCheckConst.WRITE_ABLE, must_exist=False)

    if os.path.isfile(args.golden_path):
        logger.warning('Comparison between ATB data files is not yet supported')
        return

    golden_stats_path = check_and_get_real_path(os.path.join(args.golden_path, 'statistic.csv'),
                                                FileCheckConst.READ_ABLE, file_type=FileCheckConst.CSV_SUFFIX)
    target_stats_path = check_and_get_real_path(os.path.join(args.target_path, 'statistic.csv'),
                                                FileCheckConst.READ_ABLE, file_type=FileCheckConst.CSV_SUFFIX)
    golden_csv_content = read_csv(golden_stats_path, as_pd=False)
    target_csv_content = read_csv(target_stats_path, as_pd=False)
    golden_comparison_mode = get_comparison_mode(golden_csv_content)
    target_comparison_mode = get_comparison_mode(target_csv_content)
    if (golden_comparison_mode != target_comparison_mode or
       golden_comparison_mode == CompareConst.UNKNOWN_COMPARISION_MODE):
        logger.error('Unvalid ATB dump data')
        return

    if golden_comparison_mode == CompareConst.TENSOR_COMPARISION_MODE and not torch_available:
        logger.error('Unable to compare ATB Tensor without torch. Please install with \"pip install torch\"')
        return

    golden_stats_map = get_stats_map(golden_csv_content, args.golden_path)
    target_stats_map = get_stats_map(target_csv_content, args.target_path)
    comparison_result_map = {}
    for col_name in COMMON_HEADER:
        comparison_result_map[f'Target {col_name}'] = []
        comparison_result_map[f'Golden {col_name}'] = []

    if golden_comparison_mode == CompareConst.STATS_COMPARISION_MODE:
        for col_name in STATS_SPECIAL_HEADER:
            comparison_result_map[col_name] = []
        for col_name in COMMON_HEADER_STATS:
            comparison_result_map[f'Target {col_name}'] = []
            comparison_result_map[f'Golden {col_name}'] = []
        cal_comparison_metrics_by_stats(golden_stats_map, target_stats_map, comparison_result_map)
    else:
        for col_name in TENSOR_SPECIAL_HEADER:
            comparison_result_map[col_name] = []
        for col_name in COMMON_HEADER_STATS:
            comparison_result_map[f'Target {col_name}'] = []
            comparison_result_map[f'Golden {col_name}'] = []
        cal_comparison_metrics_by_tensor(golden_stats_map, target_stats_map, comparison_result_map)

    if comparison_result_map.get(f'Target {CompareConst.ATB_DATA_NAME}'):
        create_directory(args.output_path)
        file_name = (f'atb_{golden_comparison_mode}_' +
                     f'compare_result_{time.strftime("%Y%m%d%H%M%S", time.localtime(time.time()))}.xlsx')
        file_path = os.path.join(args.output_path, file_name)
        save_excel(file_path, pd.DataFrame(comparison_result_map))
        logger.info(f'Complete ATB dump data comparison. Comparison result has been saved at {file_path}')


def get_comparison_mode(csv_content):
    if len(csv_content) < MIN_ROW_COUNT_OF_CSV or not csv_content[ROW_NUMBER_OF_FIRST_STAT]:
        return CompareConst.UNKNOWN_COMPARISION_MODE

    if csv_content[ROW_NUMBER_OF_FIRST_STAT][-1] == CompareConst.N_A:
        return CompareConst.STATS_COMPARISION_MODE

    return CompareConst.TENSOR_COMPARISION_MODE


def get_stats_map(csv_content, root_path):
    stats_map = dict()

    if len(csv_content) < MIN_ROW_COUNT_OF_CSV:
        return stats_map

    csv_content.pop(0)

    for tensor_stat in csv_content:
        if len(tensor_stat) != COL_COUNT_OF_CSV:
            continue

        single_names = tensor_stat[OP_NAME_COL_NUM].split(Const.SCOPE_SEPARATOR)
        single_ids = tensor_stat[OP_ID_COL_NUM].split(Const.REPLACEMENT_CHARACTER)
        if (len(single_names) != len(single_ids)):
            continue
        data_name = ''
        for index, name in enumerate(single_names):
            data_name += f'{single_ids[index]}_{name}{Const.SCOPE_SEPARATOR}'
        data_name += f'{tensor_stat[IO_TYPE_COL_NUM]}.{tensor_stat[TENSOR_INDEX_COL_NUM]}'

        tensor_path = tensor_stat[TENSOR_PATH_COL_NUM]
        if tensor_path != CompareConst.N_A:
            tensor_path = tensor_path[tensor_path.rfind('atb_dump_data'):]
            tensor_path_vec = tensor_path.split(Const.SCOPE_SEPARATOR)
            if len(tensor_path_vec) < 5:
                continue
            tensor_path = os.path.join(root_path, Const.SCOPE_SEPARATOR.join(tensor_path_vec[4:]))

        single_stat_map = {
            CompareConst.ATB_DATA_NAME: data_name,
            CompareConst.DEVICE_AND_PID: tensor_stat[DEVICE_AND_PID_COL_NUM],
            CompareConst.EXECUTION_COUNT: tensor_stat[EXECUTION_COUNT_COL_NUM],
            CompareConst.DATA_TYPE: tensor_stat[DTYPE_COL_NUM],
            CompareConst.DATA_SHAPE: tensor_stat[SHAPE_COL_NUM],
            Const.MAX: tensor_stat[MAX_VALUE_COL_NUM],
            Const.MIN: tensor_stat[MIN_VALUE_COL_NUM],
            Const.MEAN: tensor_stat[MEAN_VALUE_COL_NUM],
            Const.NORM: tensor_stat[NORM_VALUE_COL_NUM],
            CompareConst.TENSOR_PATH: tensor_path
        }

        stats_map[data_name] = single_stat_map

    return stats_map


def convert_str_to_float(float_str):
    float_value = 0
    try:
        float_value = float(float_str)
    except ValueError:
        return False, float_value

    return True, float_value


def convert_str_to_int(int_str, default_value=0):
    int_value = default_value
    try:
        int_value = int(int_str)
    except ValueError:
        return False, int_value

    return True, int_value


def cal_comparison_metrics_by_stats(golden_stats_map, target_stats_map, comparison_result_map):
    for data_name in target_stats_map:
        if data_name not in golden_stats_map:
            continue

        for col_name in COMMON_HEADER:
            comparison_result_map[f'Target {col_name}'].append(target_stats_map[data_name][col_name])
            comparison_result_map[f'Golden {col_name}'].append(golden_stats_map[data_name][col_name])

        for metric in Const.SUMMARY_METRICS_LIST:
            comparison_result_map[f'Target {metric}'].append(target_stats_map[data_name][metric])
            comparison_result_map[f'Golden {metric}'].append(golden_stats_map[data_name][metric])
            target_crt_ret, target_crt_value = convert_str_to_float(target_stats_map[data_name][metric])
            golden_crt_ret, golden_crt_value = convert_str_to_float(golden_stats_map[data_name][metric])
            if not (target_crt_ret and golden_crt_ret):
                comparison_result_map[f'{metric} Diff'].append(CompareConst.N_A)
                comparison_result_map[f'Relative Err of {metric}(%)'].append(CompareConst.N_A)
                continue

            diff_value = target_crt_value - golden_crt_value
            comparison_result_map[f'{metric} Diff'].append(str(diff_value))
            if diff_value == 0:
                comparison_result_map[f'Relative Err of {metric}(%)'].append('0.0')
                continue
            if golden_crt_value == 0:
                comparison_result_map[f'Relative Err of {metric}(%)'].append(CompareConst.N_A)
                continue
            comparison_result_map[f'Relative Err of {metric}(%)'].append(
                str(abs(diff_value / golden_crt_value * 100)))

    return comparison_result_map


def cal_comparison_metrics_by_tensor(golden_stats_map, target_stats_map, comparison_result_map):
    for data_name in target_stats_map:
        if data_name not in golden_stats_map:
            continue

        for col_name in COMMON_HEADER:
            comparison_result_map[f'Target {col_name}'].append(target_stats_map[data_name][col_name])
            comparison_result_map[f'Golden {col_name}'].append(golden_stats_map[data_name][col_name])

        target_bin = TensorBinFile(target_stats_map[data_name][CompareConst.TENSOR_PATH])
        golden_bin = TensorBinFile(golden_stats_map[data_name][CompareConst.TENSOR_PATH])
        target_tensor = target_bin.get_data()
        golden_tensor = golden_bin.get_data()
        target_tensor_stats = cal_single_tensor_stats(target_tensor, target_bin.is_valid)
        golden_tensor_stats = cal_single_tensor_stats(golden_tensor, golden_bin.is_valid)

        for col_name in COMMON_HEADER_STATS:
            comparison_result_map[f'Target {col_name}'].append(target_tensor_stats[col_name])
            comparison_result_map[f'Golden {col_name}'].append(golden_tensor_stats[col_name])

        comparison_metrics = [CompareConst.N_A] * len(TENSOR_SPECIAL_HEADER)
        if target_bin.is_valid and golden_bin.is_valid:
            comparison_metrics = compare_single_tensor(golden_tensor, target_tensor)

        if len(comparison_metrics) < len(TENSOR_SPECIAL_HEADER):
            comparison_metrics += [CompareConst.N_A] * (len(TENSOR_SPECIAL_HEADER) - len(comparison_metrics))
        for index, col_name in enumerate(TENSOR_SPECIAL_HEADER):
            comparison_result_map[col_name].append(comparison_metrics[index])

    return comparison_result_map


def cal_single_tensor_stats(tensor, is_valid=True):
    tensor_stats = {}
    for stat in COMMON_HEADER_STATS:
        tensor_stats[stat] = CompareConst.N_A

    if not is_valid or not tensor.numel() or not tensor.shape:
        return tensor_stats

    if tensor.dtype == torch.bool:
        tensor_stats[Const.MAX] = str(torch.any(tensor))
        tensor_stats[Const.MIN] = str(torch.all(tensor))
        return tensor_stats

    if not torch.is_complex(tensor) and (tensor.dtype == torch.float64 or not tensor.is_floating_point()):
        tensor = tensor.float()

    max_value = CompareConst.N_A
    min_value = CompareConst.N_A

    if not torch.is_complex(tensor):
        max_value = str(tensor.max().item())
        min_value = str(tensor.min().item())
    mean_value = str(tensor.mean().item())
    norm_value = str(tensor.norm().item())
    tensor_stats[Const.MAX] = CompareConst.N_A if max_value == 'nan' else max_value
    tensor_stats[Const.MIN] = CompareConst.N_A if min_value == 'nan' else min_value
    tensor_stats[Const.MEAN] = CompareConst.N_A if mean_value == 'nan' else mean_value
    tensor_stats[Const.NORM] = CompareConst.N_A if norm_value == 'nan' else norm_value

    return tensor_stats


def compare_single_tensor(golden_tensor, target_tensor):
    result_list = [CompareConst.N_A] * len(TENSOR_SPECIAL_HEADER)

    if golden_tensor.dtype == torch.bfloat16:
        golden_tensor = golden_tensor.to(torch.float32)
    golden_value = golden_tensor.numpy()
    if target_tensor.dtype == torch.bfloat16:
        target_tensor = target_tensor.to(torch.float32)
    target_value = target_tensor.numpy()

    target_value, golden_value, error_flag, _ = get_error_flag_and_msg(target_value, golden_value)
    if error_flag:
        return result_list

    result_list, _ = compare_ops_apply(target_value, golden_value, False, '')
    result_list = [str(ret) if str(ret) != CompareConst.NAN else CompareConst.N_A for ret in result_list]
    return result_list


class TensorBinFile:
    def __init__(self, file_path) -> None:
        self.file_path = file_path
        self.real_data_path = ''
        self.dtype = 0
        self.format = 0
        self.dims = []
        self.dtype_dict = {
            0: torch.float32,
            1: torch.float16,
            2: torch.int8,
            3: torch.int32,
            9: torch.int64,
            12: torch.bool,
            27: torch.bfloat16
        }
        self.is_valid = True
        self._parse_bin_file()

    def get_data(self):
        tensor = torch.zeros(1)
        if not self.is_valid or self.real_data_path:
            return tensor
        if self.dtype not in self.dtype_dict:
            logger.warning(f'Unsupported dtype: {self.dtype}')
            self.is_valid = False
            return tensor
        dtype = self.dtype_dict.get(self.dtype)
        try:
            tensor = torch.frombuffer(array.array('b', self.obj_buffer), dtype=dtype)
            tensor = tensor.view(self.dims)
        except Exception:
            logger.warning(f'Can not convert {self.file_path} to PyTorch Tensor')
            self.is_valid = False
            return torch.zeros(1)
        return tensor

    def _parse_bin_file(self):
        try:
            file_data = get_file_content_bytes(self.file_path)
        except FileCheckException:
            self.is_valid = False
            return

        begin_offset = 0
        for i, byte in enumerate(file_data):
            if byte != ord("\n"):
                continue
            line = file_data[begin_offset: i].decode("utf-8")
            begin_offset = i + 1
            fields = line.split("=")
            if len(fields) != 2:
                self.is_valid = False
                return

            attr_name = fields[0]
            attr_value = fields[1]
            if attr_name == ATTR_END:
                self.obj_buffer = file_data[i + 1:]
                return

            if not attr_name.startswith("$"):
                self._parse_user_attr(attr_name, attr_value)

    def _parse_user_attr(self, attr_name, attr_value):
        ret = True
        if attr_name == "data":
            self.real_data_path = attr_value
            ret = False
        elif attr_name == "dtype":
            ret, self.dtype = convert_str_to_int(attr_value, -1)
        elif attr_name == "format":
            ret, self.format = convert_str_to_int(attr_value, -1)
        elif attr_name == "dims":
            self.dims = attr_value.split(",")
            self.dims = [convert_str_to_int(dim, -1)[1] for dim in self.dims]
            if -1 in self.dims:
                ret = False
        if not ret:
            self.is_valid = False
